# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for helper functions for IO with HDF5 files """

import os
import ast
from warnings import warn
import h5py


WRITE_HDF5 = "write_hdf5"
READ_HDF5 = "read_hdf5"
REQUIRED_POSITIONAL_ARGUMENT = "required positional argument"
GET_IDENTIFYING_ATTRIBUTES = "get_identifying_attributes"

ERROR_GROUP_NAMES = "Either provide prefix for default group name or full group name"
ERROR_WRITE       = "Object of type '{}' does not have a write_hdf5 method"
ERROR_READ        = "Class '{}' does not have a read_hdf5 method"
ERROR_IO          = "IO is not implemented correctly"
ERROR_IDENTIFIER  = "Either give identifier '{}' of the class or a full group name"
ERROR_GROUP       = "Did not find group '{}' in HDF5 file '{}'"
ERROR_SUPER       = "(Super?) Class does not have a {} method"

WARNING_EXISTED   = "Group '{}' already existed in file '{}' and got deleted"
WARNING_PARSE     = "Intermediate identifier is None, group name cannot be parsed back"
WARNING_STRING    = "Experimental parsing of attributes from group name"


def save(obj, filename, prefix_group_name=None, full_group_name=None, mode="a"):
    """
    save object in an HDF5 file

    :param obj: the object to be saved
    :param (str) filename: the name of the HDF5 file
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (str or None) full_group_name: the full name of the HDF5 group inside the HDF5 file replacing the default,
                                          if the group name already exists in the file, it will be overwritten
    :param (str) mode: append to file ('a') or overwrite file ('w')
    """
    if prefix_group_name and full_group_name:
        raise ValueError(ERROR_GROUP_NAMES)

    if not full_group_name:
        identifiers = {identifier: getattr(obj, identifier) for identifier in get_identifying_attributes(obj)}
        _check_nones(list(identifiers.values()))
        group_name = get_group_name(type(obj), prefix_group_name, **identifiers)
    else:
        group_name = full_group_name

    with h5py.File(filename, mode) as h5file:
        save_in(obj, h5file, group_name)

def save_in(obj, h5_file_or_group, group_name=None):
    """
    save object in an already opened HDF5 file or a group in an HDF5 file

    :param obj: the object to be saved
    :param (h5py.File or h5py.Group) h5_file_or_group: the HDF5 file or the group
    :param (str or None) group_name: the name of the HDF5 group inside the given HDF5 file or group,
                                     if the group name already exists in the file it will be overwritten,
                                     if not given, no subgroup is created
    :raises ValueError if the object does not have a 'write_hdf5' method
    """
    if group_name:
        # if the group is already there, it is deleted and recreated
        if group_name in h5_file_or_group:
            del h5_file_or_group[group_name]
            warn(WARNING_EXISTED.format(group_name, h5_file_or_group.filename))
        group = h5_file_or_group.create_group(group_name)
    else:
        group = h5_file_or_group

    _try_write(obj, group)

def _try_write(obj, group):
    """
    try to call 'write_hdf5' on the object and raise error if it does not have it
    (this is either because the class does not have an IO implementation or the IO is not loaded properly)
    """
    try:
        obj.write_hdf5(group)
    except AttributeError as ae:
        # checking that the error is about the correct attribute
        assert WRITE_HDF5 in ae.args[0]
        raise ValueError(ERROR_WRITE.format(type(obj).__name__)) from ae
    except ValueError as ve:
        # Since the IO might be nested (some objects' IO might rely on the IO of another)
        # there might already be an ValueError thrown
        # catching it and re-raising it for better readability of the stack trace in console
        raise ve from ve

def _check_nones(values):
    if None in values:
        for i in range(len(values) - 1):
            if values[i] is None and values[i+1] is not None:
                warn(WARNING_PARSE)
                return


def load(cls, filename, prefix_group_name=None, full_group_name=None, **identifiers):
    """
    load object of a certain class from an HDF5 file

    :param cls: the type of the object to be loaded
    :param (str) filename: the name of HDF5 file
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (str or None) full_group_name: the full name of the HDF5 group inside the HDF5 file replacing the default
    :param identifiers: keywords dict with {attribute_name : value} which are unique for an object of the given class
    :return the loaded object of the given class
    """
    if prefix_group_name and full_group_name:
        raise ValueError(ERROR_GROUP_NAMES)

    if full_group_name and not identifiers:
        # this is possible but should be avoided if identifiers can be given explicitly
        identifiers = parse_identifiers_from_group_name(cls, full_group_name)

    # get the group in the file
    group_name = full_group_name or get_group_name(cls, prefix_group_name, **identifiers)

    with h5py.File(filename, "r") as h5file:
        loaded_obj = load_from(cls, h5file, group_name=group_name, **identifiers)
    return loaded_obj

def load_from(cls, h5_file_or_group, group_name=None, **identifiers):
    """
    load object of a certain class from an already opened HDF5 file or a group in an HDF5 file

    :param cls: the type of the object to be loaded
    :param (h5py.File or h5py.Group) h5_file_or_group: the file or the group
    :param (str or None) group_name: the name of the HDF5 group inside the given HDF5 file or group,
                                     if the group name already exists in the file it will be overwritten,
                                     if not given, no subgroup is used
    :param identifiers: keywords dict with {attribute_name : value} which are unique for an object of the given class
    :raises ValueError if the object does not have a 'read_hdf5' method
    :return the loaded object of the given class
    """
    loaded_data = load_data_from(cls, h5_file_or_group, group_name=group_name, **identifiers)
    try:
        # initialize object with loaded data
        obj = cls(**loaded_data)
    except TypeError as te:
        assert REQUIRED_POSITIONAL_ARGUMENT in te.args[0]
        missing = te.args[0].split("'")[1]
        assert missing in get_identifying_attributes(cls), ERROR_IO
        raise ValueError(ERROR_IDENTIFIER.format(missing)) from te
    return obj

def load_data_from(cls, h5_file_or_group, group_name=None, **identifiers):
    """
    load data of object of a certain class from an already opened HDF5 file or a group in an HDF5 file

    :param cls: the type of object to be loaded
    :param (h5py.File or h5py.Group) h5_file_or_group: the file or the group
    :param (str or None) group_name: the name of the HDF5 group inside the given HDF5 file or group,
                                     if the group name already exists in the file it will be overwritten,
                                     if not given, no subgroup is used
    :param identifiers: keywords dict with {attribute_name : value} which are unique for an object of the given class
    :raises ValueError if the object does not have a 'read_hdf5' method
    :return the loaded data for an object of the given class
    """
    if group_name:
        if group_name not in h5_file_or_group:
            raise ValueError(ERROR_GROUP.format(group_name, h5_file_or_group.file.filename))
        group = h5_file_or_group[group_name]
    else:
        group = h5_file_or_group

    if group_name and not identifiers:
        # this is possible but should be avoided if identifiers can be given explicitly
        identifiers = parse_identifiers_from_group_name(cls, group_name)

    init_kwargs = _try_read(cls, group)
    init_kwargs.update(**identifiers)
    return init_kwargs

def _try_read(cls, group):
    """
    try to call 'read_hdf5' on the class and raise error if it does not have it
    (this is either because the class does not have an IO implementation or the IO is not loaded properly)
    """
    try:
        loaded_data = cls.read_hdf5(group)
    except AttributeError as ae:
        # checking that the error is about the correct attribute
        assert READ_HDF5 in ae.args[0]
        if hasattr(cls, "__name__"):
            raise ValueError(ERROR_READ.format(cls.__name__)) from ae
        raise ValueError(ERROR_SUPER.format(READ_HDF5)) from ae

    except ValueError as ve:
        # Since the IO might be nested (some objects' IO might rely on the IO of another)
        # there might already be an ValueError thrown
        # catching it and re-raising it for better readability of the stack trace in console
        raise ve from ve
    return loaded_data



def load_all(cls, filename, prefix_group_name=None, used_default=True, index_range=None):
    """
    load all objects of a certain class from the HDF5 file

    :param cls: the type of the objects to be loaded
    :param (str) filename: the name of the HDF5 file
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (bool) used_default: if True, the default method was used for storing,
                                you can set this flag to False and no default class prefix will be added
                                but this could cause problems if there are different objects stored in the group
    :param (list or range or None) index_range: specify the objects which shall be loaded according to the index range
                                                they appear in
    :return: the loaded objects of the given type
    """
    with h5py.File(filename, "r") as h5file:
        loaded_objs = load_all_from(cls, h5file, prefix_group_name, used_default, index_range)
    return loaded_objs

def load_all_from(cls, h5_file_or_group, prefix_group_name=None, used_default=True, index_range=None):
    """
    load all objects of a certain class from an already opened HDF5 file or a group in an HDF5 file

    :param cls: the type of the objects to be loaded
    :param (h5py.File or h5py.Group) h5_file_or_group: the file or the group
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (bool) used_default: if True, the default method was used for storing,
                                you can set this flag to False and no default class prefix will be added
                                but this could cause problems if there are different objects stored in the group
    :param (list or range or None) index_range: specify the objects which shall be loaded according to the index range
                                                they appear in
    :return: the loaded objects of the given type
    """
    depth = 0
    if used_default:
        prefix_group_name = (prefix_group_name or "") + "/" + cls.__name__
        depth = len(get_identifying_attributes(cls)) - 1
    all_subgroup_names = get_all_subgroup_names_from(h5_file_or_group, prefix_group_name, depth)
    if index_range:
        all_subgroup_names = [all_subgroup_names[i] for i in index_range]
    all_objs = []
    for name in all_subgroup_names:
        obj = load_from(cls, h5_file_or_group, name)
        all_objs.append(obj)
    return all_objs

def get_all_subgroup_names(filename, prefix_group_name=None, depth=0, complete=True):
    """
    get the names of all objects in the group in the HDF5 file

    :param (str) filename: the name of the HDF5 file
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (int) depth: the number of subgroups to collect recursively
    :param (bool) complete: if True, the full group names are returned
    :return: the names of all objects stored in the file
    """
    with h5py.File(filename, "r") as h5file:
        names = get_all_subgroup_names_from(h5file, prefix_group_name, depth, complete)
    return names

def get_all_subgroup_names_from(h5_file_or_group, prefix_group_name=None, depth=0, complete=True):
    """
    get the names of all objects in an already opened HDF5 file or a group in an HDF5 file

    :param (h5py.File or h5py.Group) h5_file_or_group: the file or the group
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (int) depth: the number of subgroups to collect recursively
    :param (bool) complete: if True, the full group names are returned
    :return: the names of all objects stored in the file
    """
    if prefix_group_name and not prefix_group_name in h5_file_or_group:
        return []
    if not prefix_group_name:
        group = h5_file_or_group
    else:
        group = h5_file_or_group[prefix_group_name]

    subgroup_names = _get_subgroup_names_recursive(group, depth)
    if complete and prefix_group_name:
        if not str(prefix_group_name).endswith("/"):
            prefix_group_name = prefix_group_name + "/"
        subgroup_names = sorted([prefix_group_name + str(name) for name in subgroup_names])

    return subgroup_names

def _get_subgroup_names_recursive(group, depth=0):
    """ go through all subdirectories recursively """
    if depth == 0:
        return sorted([str(name) for name in group.keys()])
    group_names = []
    for subgroup_name in group.keys():
        subgroup = group[subgroup_name]
        names = sorted([subgroup_name + "/" + str(name) for name in _get_subgroup_names_recursive(subgroup, depth-1)])
        group_names += names
    return group_names

def is_subgroup(h5_file_or_group, name):
    """
    check whether the given name is a group in the opened HDF5 file or a group in an HDF5 file
    (in contrast to e.g. a dataset)

    :param (h5py.File or h5py.Group) h5_file_or_group: the file or the group
    :param (str) name: the name of the group to be checked
    """
    if name in h5_file_or_group:
        sub_obj = h5_file_or_group[name]
        return isinstance(sub_obj, h5py.Group)
    return False


def exists(cls, filename, prefix_group_name=None, full_group_name=None, **identifiers):
    """
    check if an object of a certain class already exists in an HDF5 file

    :param cls: the type of the object to be checked
    :param (str) filename: the name of HDF5 file
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param (str or None) full_group_name: the full name of the HDF5 group inside the HDF5 file replacing the default
    :param identifiers: keywords dict with {attribute_name : value} which are unique for an object of the given class
    :return: True if there is already an object of the given type in the file
    """
    if prefix_group_name and full_group_name:
        raise ValueError(ERROR_GROUP_NAMES)

    if os.path.exists(filename):
        with h5py.File(filename, "r") as h5file:
            full_group_name = full_group_name or get_group_name(cls, prefix_group_name, **identifiers)
            return full_group_name in h5file
    return False

def get_identifying_attributes(cls):
    """
    get the identifiers of the class, if it implements them

    :param cls: the type of the class to be checked
    :return: the list of identifying attributes
    """
    if hasattr(cls, GET_IDENTIFYING_ATTRIBUTES) and callable(cls.get_identifying_attributes):
        return cls.get_identifying_attributes()
    return []

def get_group_name(cls, prefix_group_name=None, **identifiers):
    """
    get the default group name from the class

    :param cls: the class to get the group name from
    :param (str or None) prefix_group_name: the group hierarchy above the default group name
    :param identifiers: keywords dict with {attribute_name : value} which are unique for an object of the given class
    :return: the group name
    """
    group_name = (prefix_group_name + "/" if prefix_group_name else "") + cls.__name__
    expected_attributes = get_identifying_attributes(cls)
    for identifier in expected_attributes:
        if identifier in identifiers and not identifiers[identifier] is None:
            group_name += f"/{identifiers[identifier]}"
    return group_name

def parse_identifiers_from_group_name(cls, group_name):
    """
    parse the identifying attributes from the group name (experimental status),
    starts at end of group name string with last identifier and goes backwards in the group hierarchy

    :param cls: the class to get the identifiers for
    :param (str) group_name: the name of the group under which the object is stored
    :return: keywords dict with {attribute_name : value} which are unique for an object of the given class
    """
    sorted_identifiers = get_identifying_attributes(cls)
    subgroups = group_name.split("/")
    if subgroups[0] == cls.__name__:
        # in case of the default group name it starts with the class name, which we can drop here
        subgroups = subgroups[1:]
    # if we have less subgroups than identifiers then we only take the first identifiers
    sorted_identifiers = sorted_identifiers[:len(subgroups)][::-1]
    # trying to parse the subgroup name to the original values starting from last
    identifiers = {identifier : _try_parse(subgroups[-(i + 1)]) for i, identifier in enumerate(sorted_identifiers)}
    return identifiers

def _try_parse(value):
    """ try to parse the original value from the string """
    warn(WARNING_STRING)
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError):
        return value



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for ExampleConfig """

from quark.utils import Config


class ExampleConfig(Config):
    """ example implementation of a configuration for testing """

    def __init__(self,
                 float_option=None,
                 int_option=None,
                 str_option=None,
                 list_option=None,
                 dict_option=None,
                 set_defaults=False):
        """
        initialize ExampleConfig object

        :param float_option: option with float
        :param int_option: option with int
        :param str_option: option with str
        :param list_option: option with list
        :param dict_option: option with dict
        :param set_defaults: if True, set options which are not given explicitly to the default values
        """
        super().__init__(set_defaults,
                         float_option=float_option,
                         int_option=int_option,
                         list_option=list_option,
                         str_option=str_option,
                         dict_option=dict_option)

    @staticmethod
    def get_options_dict():
        """ get the dictionary of all possible options and additional information such as their default values """
        return {"float_option" : {"type": float, "default": 20.0,           "help": 'Just a float for testing'},
                "int_option"   : {"type": int,   "default": 10,             "help": 'Just an int for testing'},
                "str_option"   : {"type": str,   "default": "some_string",  "help": 'Just a str for testing'},
                "list_option"  : {"type": list,  "default": None,           "help": 'Just a list for testing'},
                "dict_option"  : {"type": dict,  "default": {"default": 0}, "help": 'Just a dict for testing'}}



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for Polynomial """

from functools import cached_property
from itertools import product
from numbers import Real, Integral
from warnings import warn
from typing_extensions import Self
import numpy as np
from pyscipopt import Expr
import networkx as nx

from .utils import variables


DEGREE             = "degree"
VARIABLES          = "variables"
QUADRATIC          = "quadratic"
SIGMAS             = "sigmas"
COEFFICIENTS_LISTS = "coefficients_lists"
COEFFICIENTS_INFO  = "coefficients_info"
EXPECTED           = "Expected variable type"
PREPROCESS_RULE    = "preprocess_rule"
LINEAR             = "linear"
MIN_ABS_COEFF      = "min_abs_coeff"
MAX_ABS_COEFF      = "max_abs_coeff"
MIN_DIST           = "min_dist"
MIN_ABS_DIST       = "min_abs_dist"
MAX_MIN_RATIO      = "max_min_ratio"
MAX_DIST_RATIO     = "max_dist_ratio"
MAX_ABS_DIST_RATIO = "max_abs_dist_ratio"
VALUES_INFO_KWARGS = [MIN_ABS_COEFF, MAX_ABS_COEFF, MIN_DIST, MIN_ABS_DIST,
                     MAX_MIN_RATIO, MAX_DIST_RATIO, MAX_ABS_DIST_RATIO]

ERROR_COEFF      = "Invalid coefficient '{}'"
ERROR_IMMUTABLE  = "Cannot modify immutable Polynomial object"
ERROR_ADD        = "Can only add numbers or polynomials"
ERROR_MULTIPLY   = "Can only multiply with numbers or polynomials"
ERROR_EXPONENT   = "Exponent should be integer"
ERROR_DIVISOR    = "Divisor should be number"
ERROR_ASSIGN     = "Can only assign numbers or polynomials"
ERROR_REPLACE    = "Replacement is invalid, the resulting polynomial has different variable types"
ERROR_PREPROCESS = "No preprocess rule implemented, needs to be provided"
ERROR_QUADRATIC  = "Only applicable for quadratic polynomials"
ERROR_FLAT       = "Only applicable for flat polynomials"
ERROR_COMPACT    = "Either provide compact polynomial or the total number of variables"
ERROR_MATRIX     = "Can only process matrices with at most two dimensions"
ERROR_DOUBLE     = "Double exponents are not supported"
ERROR_KEPT       = "Kept variables and replaced variables are not compatible"
ERROR_SORTING    = "Sorting failed, this is most likely due to variables with the same prefix but different " \
                   "structure, as e.g. `('x', 1)` and `('x', 'a')`"
ERROR_DEGREE     = "Provide degree greater than 0"

WARNING_CAST        = "{} will be cast to Polynomial when added or multiplied from the right"
WARNING_CONSTANT    = "Constant offset of {} is dropped"


# pylint: disable=too-many-public-methods
class Polynomial(dict):
    """
    A polynomial is defined as the sum of weighted monomials, like

        P = P0
            * +sum_{i} P1_{i} x_i
            * +sum_{i, j} P2_{i, j} x_i x_j
            * +sum_{i, j, k} P3_{i, j, k} x_i x_j x_k
            * +...,

    where x_i are variables.

    In particular, P is stored as a Python dictionary mapping the tuple of variables, e.g. here indices,
    to the value of the coefficient:

        * var_tuple --> coefficient
        * `()`      --> P0            (offset)
        * `(i,)`    --> P1_{i}        (linear monomials)
        * `(i, j)`  --> P2_{i, j}     (quadratic monomials)
        * ...

    The variables in the tuples of the polynomial can be either of type

        * ints              (e.g. `(1, 2)`),
        * strings           (e.g. `('x', 'y')`),
        * tuples themselves (e.g. `(('x', 1), ('x', 2))`)     (only containing ints and strings),

    but the cases cannot be mixed up, since they cannot be compared with each other for sorting
    (therefore `(('x', 1), 'y')` is for instance not possible).

    The type of the variables can be unset if the Polynomial is empty or just a constant.
    """

    def __init__(self, d=None, variable_tuples_already_formatted=False):
        """
        initialize Polynomial object

        :param (dict) d: mapping from tuples of variables to coefficients
        :param (bool) variable_tuples_already_formatted: if True, d will be used as passed to save time on sorting,
                                                         use if d is, e.g., a dictionary of an existing Polynomial
                                                         (be careful with other cases, this avoids also type checks...),
                                                         otherwise every key in d will be sorted
        """
        self.var_type = None  # type of the variables

        if isinstance(d, self.__class__) or variable_tuples_already_formatted:
            super().__init__(d)
            some_key = next((k for k in self.keys() if k), None)
            self.var_type = variables.get_common_type(some_key)
            return

        if self._is_valid_coefficient(d):
            # just a constant was passed
            super().__init__({(): d})
            return

        d = d or {}
        poly = {}
        for var_tuple, coeff in d.items():
            if var_tuple and not self.var_type:
                self.var_type = variables.get_common_type(var_tuple)
            var_tuple = self._format_variable_tuple(var_tuple)
            if not self._is_valid_coefficient(coeff):
                raise ValueError(ERROR_COEFF.format(coeff))
            _add_monomial(poly, var_tuple, coeff)
        super().__init__(poly)

    def _format_variable_tuple(self, var_tuple):
        """
        transform the tuple of variables into the standard format, also checks the type of the variables

        :param (tuple or int or str) var_tuple: unordered tuple of variables
        :return: the formatted and ordered tuple of variables
        """
        #  if there is a single int or string, it should be a tuple of length one
        if not isinstance(var_tuple, tuple):
            var_tuple = (var_tuple,)
        # check if all variables in the tuple have the same preset type
        for var in var_tuple:
            variables.check_type_against(var, self.var_type)
        var_tuple = tuple(_try_sorted(var_tuple))
        return var_tuple

    @staticmethod
    def _is_valid_coefficient(coeff):
        """ check if the coefficient has the correct type (just numbers) """
        return isinstance(coeff, Real)

    @staticmethod
    def _immutable(*args, **kwargs):
        """ workaround to make some inherited functions unusable """
        raise TypeError(ERROR_IMMUTABLE)

    __setitem__ = _immutable
    __delitem__ = _immutable
    pop = _immutable
    popitem = _immutable
    clear = _immutable
    update = _immutable
    setdefault = _immutable

    # for caching of computed values

    @cached_property
    def degree(self):
        """ the degree of the polynomial, i.e., the largest appearing size of a monomial """
        return max((len(var_tuple) for var_tuple in self.remove_zero_coefficients()), default=0)

    @cached_property
    def variables(self):
        """ the list of variables """
        return _try_sorted(set(var for var_tuple in self for var in var_tuple))

    @cached_property
    def offset(self):
        """ the coefficient of the empty monomial """
        return self.get((), 0)

    @cached_property
    def linear_plain(self):
        """
        the weights on the variables in the linear part of the polynomial, also known as local fields

        we need to separate this method from 'Polynomial.linear' due to the representation of variables as tuples,
        which we do not want here
        """
        return {var : self.get_coefficient(var) for var in self.variables}

    @cached_property
    def linear(self):
        """ the linear part of the polynomial """
        poly = self.__class__({(var,) : value for var, value in self.linear_plain.items()},
                              variable_tuples_already_formatted=True)
        return poly.remove_zero_coefficients()

    @cached_property
    def quadratic(self):
        """ the quadratic part of the polynomial """
        quadratic_monomials = _try_sorted(var_tuple for var_tuple in self.keys() if len(var_tuple) == 2)
        return self.__class__({var_tuple : self.get_coefficient(*var_tuple) for var_tuple in quadratic_monomials},
                              variable_tuples_already_formatted=True).remove_zero_coefficients()

    @cached_property
    def coefficients_lists(self):
        """ list of lists containing the linear, quadratic, cubic, etc. coefficients """
        coefficients = [[] for _ in range(self.degree + 1)]
        for var_tuple, coeff in self.items():
            coefficients[len(var_tuple)].append(coeff)
        coeff_lists = [sorted(set(cs)) for cs in coefficients]
        return coeff_lists

    @cached_property
    def coefficients_info(self):
        """ dictionary with various coefficient information """
        # the first coefficient list is just the offset, this can be ignored
        if len(self.coefficients_lists) > 1:
            return get_info(sum(self.coefficients_lists[1:], []))
        return {}

    @cached_property
    def maximum_int_deviation(self):
        """ the largest difference from a coefficient to the closest int value """
        return max(abs(coeff - round(coeff)) for coeff_list in self.coefficients_lists for coeff in coeff_list)

    @cached_property
    def sigmas(self):
        """
        the dictionary the combined influences on a single variable, with
        variable -> (sum of all positive coupling coefficients, sum of all negative coupling coefficients)
        """
        sigmas_pos = {}
        sigmas_neg = {}
        for couple, strength in self.quadratic.items():
            for var in couple:
                if strength > 0:
                    sigmas_pos[var] = sigmas_pos.get(var, 0) + strength
                else:
                    sigmas_neg[var] = sigmas_neg.get(var, 0) + strength
        return {var : (sigmas_pos.get(var, 0), sigmas_neg.get(var, 0)) for var in self.variables}

    def __repr__(self):
        """ get nice string representation """
        self.sort_entries()
        return self.__class__.__name__ + "(" + super().__repr__() + ")"

    def __str__(self):
        """
        convert the polynomial to a human-readable string,
        non-flat polynomials might look weird and cannot be parsed back again
        """
        if self.is_constant():
            return f"{self.offset}"
        result = ''
        coeff_str = ' {:+n}'
        self.sort_entries()
        for var_tuple, coeff in self.items():
            var_str = ''
            if var_tuple:
                var_str = ' ' + ' '.join(variables.to_string(var, self.__class__) for var in var_tuple)
            result += coeff_str.format(coeff) + var_str
        return result.strip()

    def __eq__(self, other):
        """
        check equality of polynomials, monomials with zero coefficient are not considered,
        if other is a number and the polynomial is constant and has this value, we also consider them to be equal
        """
        if self._is_valid_coefficient(other):
            # we also return true also if a number is inserted and the polynomial is constant and has this value
            other = Polynomial({() : other})
        if isinstance(other, Polynomial):
            # monomials with 0 coefficient are not considered
            return super(Polynomial, self.remove_zero_coefficients()).__eq__(other.remove_zero_coefficients())
        return False

    def __ne__(self, other):
        """ check if objects are not equal, overwrites dicts not equal to use polynomials equal implementation """
        return not self.__eq__(other)

    def _warn_different_types(self, poly2):
        """ check for different types of polynomials, because of inheritance there might be more than one """
        # pylint: disable=unidiomatic-typecheck  # Check for the exact type, excluding subtypes
        if type(self) is Polynomial and not type(poly2) is Polynomial:
            poly2_type = type(poly2).__name__
            warn(WARNING_CAST.format(poly2_type), stacklevel=3)

    def __add__(self, poly2):
        """ add another polynomial or a scalar """
        result = dict(self)
        if self._is_valid_coefficient(poly2):
            _add_monomial(result, (), poly2)
        elif isinstance(poly2, Polynomial):
            self._warn_different_types(poly2)
            for var_tuple, coeff in poly2.items():
                var_tuple = self._format_variable_tuple(var_tuple)
                _add_monomial(result, var_tuple, coeff)
        else:
            try:
                return poly2.__add__(self)
            except TypeError as te:
                raise TypeError(ERROR_ADD) from te
        return self.__class__(result, variable_tuples_already_formatted=True)

    def __radd__(self, poly2):
        """ add self to another polynomial or a scalar """
        return self.__add__(poly2)

    def __sub__(self, poly2):
        """ subtract a polynomial or a scalar """
        return self.__add__(-1 * poly2)

    def __rsub__(self, poly2):
        """ subtract self from a polynomial or a scalar """
        return (-1 * self).__add__(poly2)

    def __mul__(self, poly2):
        """ multiply another polynomial or a scalar """
        if self._is_valid_coefficient(poly2):
            result = {var_tuple: coeff * poly2 for var_tuple, coeff in self.items()}
        elif isinstance(poly2, Polynomial):
            self._warn_different_types(poly2)
            result = {}
            for (vt1, c1), (vt2, c2) in product(self.items(), poly2.items()):
                var_tuple = self._format_variable_tuple(vt1 + vt2)
                _add_monomial(result, var_tuple, c1 * c2)
        else:
            try:
                return poly2.__mul__(self)
            except TypeError as te:
                raise TypeError(ERROR_MULTIPLY) from te
        return self.__class__(result, variable_tuples_already_formatted=True)

    def __rmul__(self, poly2):
        """ multiply self to another polynomial or a scalar """
        return self.__mul__(poly2)

    def __truediv__(self, other):
        """ dividing by scalar """
        if not self._is_valid_coefficient(other):
            raise TypeError(ERROR_DIVISOR)
        return self * (1 / other)

    def __pow__(self, exponent):
        """ raise to the power of the given integer exponent """
        if not isinstance(exponent, Integral):
            raise TypeError(ERROR_EXPONENT)
        result = 1
        for _ in range(exponent):
            result *= self
        return result

    def __contains__(self, variable_tuple):
        """ check whether the variables are contained as a monomial in the polynomial """
        return super().__contains__(self._format_variable_tuple(variable_tuple))

    def __getitem__(self, variable_tuple):
        return super().__getitem__(self._format_variable_tuple(variable_tuple))

    def get(self, variable_tuple, default=None):
        """ get the coefficient of the corresponding monomial in the polynomial """
        # but we catch the case where the variable monomial is given but by the formatting results in the empty tuple
        formatted_tuple = self._format_variable_tuple(variable_tuple)
        if variable_tuple and not formatted_tuple:
            return 0
        return super().get(formatted_tuple, default)

    def sort_entries(self):
        """ does not change the polynomial but the ordering of its entries in the dict, mainly for nice output """
        sorted_poly = sort_poly_dict(self)
        super().clear()
        super().update(sorted_poly)

    def copy(self) -> Self:
        """ get a copy of the polynomial """
        return self.__class__(self, variable_tuples_already_formatted=True)

    def is_constant(self):
        """ check if the polynomial contains at maximum the offset entry, hence a tuple of length 0 """
        return self.degree == 0

    def is_linear(self):
        """ check if the polynomial has a degree of at most 1 """
        return self.degree <= 1

    def is_quadratic(self):
        """ check if the polynomial has a degree of at most 2 """
        return self.degree <= 2

    def has_int_coefficients(self, precision=1e-9):
        """
        check if all coefficients of the polynomial are integers up to a certain precision

        :param (Real) precision: the deviation up to which values are considered to be integers
        """
        return self.maximum_int_deviation < precision

    def is_flat(self):
        """ check if the polynomial has only integer variables """
        return self.var_type is None or issubclass(self.var_type, Integral)

    def is_compact(self):
        """ check if the polynomial has integer variables which are consecutive starting with 0 """
        return variables.are_consecutive(self.variables)

    def get_variable_num(self):
        """ get the number of variables """
        return len(self.variables)

    def get_coefficient(self, *variables_in_monomial):
        """ get the coefficient for the tuple of variables or if not present 0 """
        return self.get(variables_in_monomial, 0)

    def get_coefficients_info_by_degree(self, degree=1):
        """
        extract the interesting information for the coefficients list of a certain degree

        :param (int) degree: the degree of the polynomial for which the coefficients info should be extracted
        :return: the coefficients info
        """
        if not degree or degree <= 0:
            raise ValueError(ERROR_DEGREE)
        if degree > len(self.coefficients_lists) - 1:
            return {}
        suffix = f"_degree_{degree}"
        return get_info(self.coefficients_lists[degree], suffix)

    def get_all_coefficients_info(self):
        """ extract the interesting information from several coefficients lists and combine the dictionaries """
        info_by_degree = {}
        for degree in range(1, len(self.coefficients_lists)):
            info_by_degree.update(self.get_coefficients_info_by_degree(degree))
        info_by_degree.update(self.coefficients_info)
        return info_by_degree

    def remove_zero_coefficients(self):
        """ remove all monomials with coefficients 0 from this polynomial """
        filtered = {var_tuple: coeff for var_tuple, coeff in self.items() if coeff != 0}
        return self.__class__(filtered, variable_tuples_already_formatted=True)

    def affine_transform(self, coefficient, offset=0):
        """
        apply the following transformation to every variable x: x -> (coefficient * x + offset)

        :param (Real) coefficient: linear coefficient of the transformation function
        :param (Real) offset: constant term of the transformation function
        :return: the new Polynomial resulting from the transformation or value if the final Polynomial is constant
        """
        transformations = {var: Polynomial({(var,): coefficient, (): offset}, variable_tuples_already_formatted=True)
                           for var in self.variables}
        return self.evaluate(transformations)

    @classmethod
    def _is_valid_var_assignment(cls, var_assignment):
        """
        check whether the given input can be assigned to the variable,
        that means it is a Polynomial, SCIP Expr or valid coefficient
        """
        return isinstance(var_assignment, (Polynomial, Expr)) or cls._is_valid_coefficient(var_assignment)

    def evaluate(self, var_assignments, keep_poly=False):
        """
        substitute all variables v with var_assignment[v], where var_assignment[v] can be either a value or Polynomial,
        the var_assignment can contain all or just a subset of the variables,
        if the Polynomial is flat, the var_assignment can also be a list with indices according to the variables

        if there is a Polynomial in the variable assignment the result will be a Polynomial,
        otherwise inherited classes will be kept (e.g. PolyIsing), if the result is constant a number will be returned

        this method can also be used to evaluate a SCIP Expr, however, in this case all variables need an assignment!

        :param (dict or list) var_assignments: mapping from variables to values or Polynomials
        :param (bool) keep_poly: if True, force to keep the result as an object of the polynomial class
                                 even if it evaluates to a constant
        :return: the new Polynomial resulting from the evaluation or value if the final Polynomial is constant
        """
        if isinstance(var_assignments, list):
            var_assignments = dict(enumerate(var_assignments))
        if any(not self._is_valid_var_assignment(va) for va in var_assignments.values()):
            raise ValueError(ERROR_ASSIGN)

        # we can only keep the originally inherited class type (e.g. PolyIsing), if we only evaluate numbers
        all_numbers = all(self._is_valid_coefficient(va) for va in var_assignments.values())
        resulting_type = self.__class__ if all_numbers else Polynomial
        try:
            result = sum(_get_evaluated_poly(*monomial, var_assignments, resulting_type) for monomial in self.items())
        except TypeError as te:
            raise ValueError(ERROR_KEPT) from te

        # check if the resulting Polynomial is just a constant, in this case only return the value, if not keep_poly
        if isinstance(result, Polynomial):
            result = result.remove_zero_coefficients()
            if result.is_constant() and not keep_poly:
                return result.get((), 0)
        if isinstance(result, Real) and keep_poly:
            return self.__class__({(): result})
        return result

    def replace_variables(self, replacement):
        """
        replace the variables in the polynomial with other variables

        :param (list or tuple or dict or function) replacement: mapping to the new variables,
                                                                if replacement is a dictionary, it should be of format
                                                                    {old_var : new_var},
                                                                can only be a list if the polynomial is flat
        :return: the new Polynomial with replaced variables
        """
        if isinstance(replacement, (list, tuple)):
            replacement = dict(enumerate(replacement))

        # here replacement of format {old_var : new_var} or function : old_var -> new_var
        result = {}
        for var_tuple, coeff in self.items():
            # e.g. from var_tuple (('x', 1), ('x', 2)) will get (0, 1) with replacement {('x', 1): 0, ('x', 2): 1}
            _add_monomial(result, variables.replace_in_tuple(var_tuple, replacement), coeff)
        try:
            return self.__class__(result)
        except TypeError as te:
            assert te.args[0].startswith(EXPECTED)
            raise ValueError(ERROR_REPLACE) from te

    def replace_variables_by_ordering(self, ordered_variables, check_all=True):
        """
        replace the variables in the polynomial by the index at which they appear in the provided list,
        if the list is complete the resulting polynomial will be compact

        :param (list) ordered_variables: list of variables of polynomial
        :param (bool) check_all: if True, checks if the given list is complete meaning all variables need a replacement,
                                 if False, only subset of variables can be replaced, they however need to have the same
                                 type as the original ones
        :return: the new Polynomial with replaced variables
        """
        if check_all:
            assert len(ordered_variables) >= self.get_variable_num()
            assert set(ordered_variables) >= set(self.variables)
        replacement = {var: i for i, var in enumerate(ordered_variables)}
        return self.replace_variables(replacement)

    def compact(self):
        """
        replace the variables in the polynomial with a consecutive indexing starting at 0,
        the indices are taken from sorting of the list of variables of the original polynomial

        :return: the new Polynomial with compact variables
        """
        return self.replace_variables_by_ordering(self.variables, False)

    def round(self, decimal_cap=None):
        """
        round the coefficients of the polynomial according to the decimal cap

        :param (int or None) decimal_cap: the number of digits of the resulting coefficients,
                                          if None, they will be ints
        :return: the new Polynomial with rounded coefficients
        """
        result = {}
        for var_tuple, coeff in self.items():
            result[var_tuple] = round(coeff, decimal_cap)
        return self.__class__(result, variable_tuples_already_formatted=True)

    def preprocess(self, preprocess_rule, unambitious=False):
        """
        simplify the polynomial according to the preprocessing rule,
        this means some variables are set to the values which are guaranteed to appear in an optimal solution

        :param (function) preprocess_rule:
            method which provides a mapping from variable sigmas and coefficient to the value to which the variable
            should be assigned, if it returns None, the variable will not be preprocessed
        :param (bool) unambitious:
            if True, only those variables will be set where their assignment is unambitious, that means, in all optimal
            solutions, they have the same value,
            if False, preprocess the variable to the straightforward value whether it might also get another value in an
            optimal solution, the objective values remains the same
        :return: the new reduced Polynomial and the assignment of the eliminated variables with variable to value
        """
        result = self.remove_zero_coefficients()

        sth_changed = True
        preprocessed_vars = {}
        while sth_changed:
            sth_changed = False
            current_pp_vars = {}
            for var in result.variables:
                coeff = result.get_coefficient(var)
                var_value = preprocess_rule(result.sigmas[var], coeff, unambitious)
                if var_value is not None:
                    current_pp_vars[var] = var_value
            if current_pp_vars:
                result = result.evaluate(current_pp_vars)
                sth_changed = not is_constant(result)
                preprocessed_vars.update(current_pp_vars)
        return result, preprocessed_vars

    def get_matrix_representation(self, num_variables=None):
        """ get the standard matrix representation of the polynomial """
        if not self.is_quadratic():
            raise ValueError(ERROR_QUADRATIC)
        if not self.is_flat():
            raise ValueError(ERROR_FLAT)
        if not self.is_compact():
            if not num_variables or num_variables < max(self.variables) + 1:
                raise ValueError(ERROR_COMPACT)
        if self.offset != 0:
            warn(WARNING_CONSTANT.format(self.offset))

        num_variables = num_variables or self.get_variable_num()
        linear = np.array([self.get_coefficient(i) for i in range(num_variables)])
        quadratic = np.array([[self.get_coefficient(i, j) if i >= j else 0
                               for i in range(num_variables)]
                              for j in range(num_variables)])
        return linear, quadratic

    @classmethod
    def get_from_matrix_representation(cls, *matrices):
        """
        get the polynomial from the standard matrix formulation

        :param matrices: arrays of different dimensions with equal size
        """
        poly = cls()
        for matrix in matrices:
            matrix = np.array(matrix)
            if len(matrix.shape) == 1:
                poly += cls(dict(enumerate(matrix)))
            if len(matrix.shape) == 2:
                poly += cls({(i, j) : val for i, ll in enumerate(matrix) for j, val in enumerate(ll)})
            if len(matrix.shape) > 2:
                raise ValueError(ERROR_MATRIX)
        return poly.remove_zero_coefficients()

    @classmethod
    def read_from_string(cls, string):
        """
        read in a polynomial from a string of the form, e.g.,
            '9 + 30.1 x1 - 25 x1^3 + 36*x2 + 60 x1 x2 + 36 x2^2 + 18 x1 x3'

        WARNING: Experimental status
        """
        # remove whitespace and multiplications
        parsed = string.replace(" ", "").replace("\n", "").replace("*", "").replace("-", "+-")
        monomials = [m for m in parsed.split('+') if m]  # take non-empty terms

        result = {}
        for m in monomials:
            factors = m.split('x')
            if factors[0] in ['', '-']:
                factors[0] = factors[0] + '1'
            coeff = float(factors[0])
            var_tuple = ()
            for f in factors[1:]:
                var_str, _, exp_str = f.partition('^')
                if exp_str.count('^'):
                    raise ValueError(ERROR_DOUBLE)
                var = int(var_str)
                exp = int(exp_str) if exp_str else 1
                var_tuple += tuple([var] * exp)
            result[var_tuple] = result.get(var_tuple, 0) + coeff
        return cls(result)

    def get_graph(self):
        """ get a graph representing the connectivity """
        if not self.is_quadratic():
            raise ValueError(ERROR_QUADRATIC)
        graph = nx.Graph()
        graph.add_nodes_from(self.linear_plain.keys())
        graph.add_edges_from(self.quadratic.keys())
        return graph


def is_constant(value_or_poly):
    """ check if the given object is either a number or a constant polynomial """
    return isinstance(value_or_poly, Real) or (isinstance(value_or_poly, Polynomial) and value_or_poly.is_constant())

def sort_poly_dict(poly_dict, reverse_monomial_degree_order=False):
    """ sort the polynomial dictionary according to the structure of the monomials """
    sorted_1 = _try_sorted(poly_dict.items(), key=lambda x: x[0])
    return dict(_try_sorted(sorted_1, key=lambda x: len(x[0]), reverse=reverse_monomial_degree_order))

def _try_sorted(iterable, key=None, reverse=False):
    """ try sorting and provide meaningful error in case it fails """
    try:
        if key:
            return sorted(iterable, key=key, reverse=reverse)
        return sorted(iterable)
    except TypeError as te:
        assert "'<' not supported" in te.args[0]
        raise ValueError(ERROR_SORTING) from te

def _add_monomial(poly_dict, var_tuple, coeff):
    """
    add a monomial (coeff * variables) to the dictionary of the polynomial

    :param (dict) poly_dict: dictionary of a polynomial with already formatted keys
    :param (tuple) var_tuple: has to be a formatted tuple of variables, can be empty for the constant term
    :param (Real) coeff: valid coefficient
    """
    poly_dict[var_tuple] = poly_dict.get(var_tuple, 0) + coeff

def _get_evaluated_poly(var_tuple, coeff, var_assignments, resulting_type):
    """ evaluate a single monomial with the given variable assignments """
    if coeff == 0:
        return 0

    term = coeff
    kept_vars = ()
    for var in var_tuple:
        if var in var_assignments.keys():
            # might have e.g. type(term) == PolyIsing and type(var_assignment[var]) == Polynomial
            # causes a warning, which we ignore, as in this case the convention is to return just a Polynomial
            # at this point we can only have either a number or a Polynomial (or a SCIP Expr)
            term *= var_assignments[var]
        else:
            kept_vars += (var,)
    if kept_vars:
        term *= resulting_type({kept_vars: 1})
    return term

def get_info(values, suffix='', precision_digits=9):
    """
    extract the interesting information from the list of values

    :param (list) values: the list of values
    :param (str) suffix: the suffix added to the keywords for differing degrees
    :param (int) precision_digits: the number of digits for the precision
    :return:
    """
    kwargs = [kw + suffix for kw in VALUES_INFO_KWARGS]
    return dict(zip(kwargs, _get_info_values(values, precision_digits)))

def _get_info_values(values, precision_digits=9):
    """
    extract the interesting values from the list of values

    :param (list) values: list of values
    :param (int) precision_digits: the number of digits for the precision
    :return: tuple (min non-zero absolute value among all values,
                    max non-zero absolute value among all values,
                    min difference between any two values,
                    min difference between any two absolute values,
                    max absolute value / min absolute value,
                    max absolute value / min differance,
                    max absolute value / min absolute difference)
    """
    if not values:
        return (0,) * 7
    precision = 10 ** (-precision_digits)

    sorted_values = sorted(set(values + [0]))
    min_dist = float('inf')
    for i in range(len(sorted_values) - 1):
        dist = sorted_values[i + 1] - sorted_values[i]
        if dist < min_dist and abs(dist) > precision:
            min_dist = dist

    sorted_absolute_values = sorted(set(abs(value) for value in sorted_values if abs(value) > precision))
    min_abs_value = sorted_absolute_values[0]
    max_abs_value = sorted_absolute_values[-1]

    min_abs_dist = min_abs_value  # distance to 0
    for i in range(len(sorted_absolute_values) - 1):
        dist = sorted_absolute_values[i + 1] - sorted_absolute_values[i]
        if dist < min_abs_dist and abs(dist) > precision:
            min_abs_dist = dist

    max_min_ratio = max_abs_value / min_abs_value
    max_dist_ratio = max_abs_value / min_dist
    max_abs_dist_ratio = max_abs_value / min_abs_dist
    values = min_abs_value, max_abs_value, min_dist, min_abs_dist, max_min_ratio, max_dist_ratio, max_abs_dist_ratio
    rounded_values = tuple(round(value, precision_digits) for value in values)
    return rounded_values



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the ConstrainedObjective """

from quark import ConstrainedObjective, ConstraintBinary
from . import add, hdf5
from .poly_utils import write_poly_with_type, read_poly_with_type


OBJECTIVE_POLY = "objective_poly"
CONSTRAINTS    = "constraints"


add.save_load_exists(ConstrainedObjective)

@add.static_method(ConstrainedObjective)
def get_identifying_attributes():
    """ the attributes that identify an object of the class """
    return ["name"]

@add.self_method(ConstrainedObjective)
def write_hdf5(constrained_objective, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (ConstrainedObjective) constrained_objective: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    write_poly_with_type(group, constrained_objective.objective_poly, OBJECTIVE_POLY)
    for name, constraint in constrained_objective.items():
        hdf5.save_in(constraint, group, name)

@add.static_method(ConstrainedObjective)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    objective_poly = read_poly_with_type(group, OBJECTIVE_POLY)
    constraints = {}
    for name in group:
        if name != OBJECTIVE_POLY:
            constraints[name] = hdf5.load_from(ConstraintBinary, group, group_name=name)
    return {OBJECTIVE_POLY: objective_poly, CONSTRAINTS: constraints}



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the ConstraintBinary """

import pytest

from quark import PolyBinary, ConstraintBinary
from quark.poly_binary import get_all_reduction_penalty_terms
from quark.constraint_binary import get_binary_representation_polynomial, get_standard_penalty_polynomial, \
    get_reduction_constraints, retrieve_reduction, matches_special_penalty_constraint, \
    get_special_constraint_penalty_poly


QUADRATIC_POLY = PolyBinary({(("x", 1),): 7, (("x", 2), ("x", 1)): 2, (): -3})
LINEAR_POLY = PolyBinary({(("x", 1),): 7, (("x", 2),): 2})
QUBIC_POLY = PolyBinary({(("x", 1), ("x", 2), ("x", 3)): 1, (("x", 1),): 1})
NAME = "test_constraint"


def test_init__check_input():
    """ test errors for inconsistent data """
    with pytest.raises(ValueError, match="At least one bound needs to be given"):
        ConstraintBinary(PolyBinary())
    with pytest.raises(ValueError, match="At the moment can only handle PolyBinary"):
        ConstraintBinary(PolyBinary().to_ising(), 0, 1)

def test_init__check_consistency():
    """ test errors for inconsistent data """
    with pytest.raises(ValueError, match="Lower bound must be less than or equal to upper bound"):
        ConstraintBinary(PolyBinary({"a" : 1}), 4, 2)
    with pytest.raises(ValueError, match="Lower bound must be less than or equal to upper bound"):
        ConstraintBinary(PolyBinary(), 4)
    with pytest.raises(ValueError, match="Lower bound must be less than or equal to upper bound"):
        ConstraintBinary(PolyBinary(), upper_bound=-4)
    with pytest.raises(ValueError, match="Unfulfillable constraint '5 == 1'"):
        ConstraintBinary(PolyBinary({(): 5}), 1, 1)
    with pytest.raises(ValueError, match="Unfulfillable constraint '4 <= 0 <= 8'"):
        ConstraintBinary(PolyBinary(), 4, 8)
    with pytest.raises(ValueError, match="Unfulfillable constraint '-6 <= 1 <= -4'"):
        ConstraintBinary(PolyBinary() + 1, upper_bound=-4, lower_bound=-6)
    with pytest.warns(UserWarning, match="Useless constraint '0 <= 5 <= 10'"):
        ConstraintBinary(PolyBinary({(): 5}), 0, 10)

    # lb <= p+(x) + p-(x) + c <= ub
    with pytest.raises(ValueError, match="Negative coefficients do not suffice to get below the upper bound"):
        ConstraintBinary(LINEAR_POLY, -1, -1)
    with pytest.raises(ValueError, match="Positive coefficients do not suffice to get above the lower bound"):
        ConstraintBinary(PolyBinary({(("x", 0),): 1}) - LINEAR_POLY, 2, 2)
    with pytest.raises(ValueError, match="Negative coefficients do not suffice to get below the upper bound"):
        ConstraintBinary(QUADRATIC_POLY, -4, -4)
    with pytest.raises(ValueError, match="Positive coefficients do not suffice to get above the lower bound"):
        ConstraintBinary(PolyBinary({(("x", 0),): 1}) - QUADRATIC_POLY, 5, 5)

    # preprocessable equality and negation
    with pytest.warns(UserWarning, match=r"The two variables 'x' and 'y' are set to be equal"):
        ConstraintBinary(PolyBinary({'x': 1, 'y': -1}), 0, 0)
    with pytest.warns(UserWarning, match=r"The two variables 'x' and 'y' are set to be negations of each other"):
        ConstraintBinary(PolyBinary({'x': 1, 'y': 1}), 1, 1)

    # contradiction found by preprocessing
    with pytest.raises(ValueError, match=r"Unfulfillable constraint '\+1 \+1 x1 \+1 x2 -1 x1 x2 == 0'"):
        ConstraintBinary(PolyBinary({"x1": 1, "x2": 1, ("x1", "x2"): -1, (): 1}), 0, 0)
    with pytest.raises(ValueError, match=r"Unfulfillable constraint '-2 \+1 x1 -1 x2 -1 x3 \+1 x2 x3 == 0'"):
        ConstraintBinary(PolyBinary({"x1": 1, "x2": -1, "x3": -1, ("x2", "x3"): 1, (): -2}), 0, 0)

    # bounds too weak on both sides
    with pytest.warns(UserWarning, match=r"Useless constraint '0 <= \+5 x1 <= 10'"):
        ConstraintBinary(PolyBinary({(1,): 5}), 0, 10)

    # bounds too weak on one sides
    with pytest.warns(UserWarning, match="Upper bound could be sharpened to 2"):
        ConstraintBinary(PolyBinary({(0,): 1, (1,): 1}), 1, 3)
    with pytest.warns(UserWarning, match="Lower bound could be sharpened to 0"):
        ConstraintBinary(PolyBinary({(0,): 1, (1,): 1}), -1, 1)

def test_init__open_bounds():
    """ test the worst case calculation of open bounds """
    constraint = ConstraintBinary(LINEAR_POLY, 1)
    assert constraint.upper_bound == 9

    constraint = ConstraintBinary(LINEAR_POLY, upper_bound=3)
    assert constraint.lower_bound == 0

    constraint = ConstraintBinary(QUADRATIC_POLY, 1)
    assert constraint.upper_bound == 6

    constraint = ConstraintBinary(QUADRATIC_POLY, upper_bound=3)
    assert constraint.lower_bound == -3

def test_add():
    """ test addition of scalars to constraint """
    exp_poly = PolyBinary({(("x", 1),): 7, (("x", 2), ("x", 1)): 2, (): -1})
    constraint = ConstraintBinary(QUADRATIC_POLY, 5, 6)
    constraint += 2
    assert constraint.polynomial == exp_poly
    assert constraint.lower_bound == 7
    assert constraint.upper_bound == 8

    constraint *= -2
    assert constraint.polynomial == -2 * exp_poly
    assert constraint.lower_bound == -16
    assert constraint.upper_bound == -14

    with pytest.raises(ValueError, match="Constant must be a number"):
        _ = ConstraintBinary(LINEAR_POLY, 0, 1) + PolyBinary()

def test_sub():
    """ test subtraction of scalars from constraint """
    exp_poly = PolyBinary({(("x", 1),): 7, (("x", 2), ("x", 1)): 2, (): -5})
    constraint = ConstraintBinary(QUADRATIC_POLY, 5, 6)
    constraint -= 2
    assert constraint.polynomial == exp_poly
    assert constraint.lower_bound == 3
    assert constraint.upper_bound == 4

    with pytest.raises(ValueError, match="Constant must be a number"):
        _ = ConstraintBinary(LINEAR_POLY, 0, 1) - PolyBinary()

def test_mul():
    """ test multiplication of scalars to constraint """
    exp_poly = PolyBinary({(("x", 1),): 14, (("x", 2), ("x", 1)): 4, (): -6})
    constraint = ConstraintBinary(QUADRATIC_POLY, 5, 6)
    constraint *= 2
    assert constraint.polynomial == exp_poly
    assert constraint.lower_bound == 10
    assert constraint.upper_bound == 12

    constraint *= -2
    assert constraint.polynomial == -2 * exp_poly
    assert constraint.lower_bound == -24
    assert constraint.upper_bound == -20

    with pytest.raises(ValueError, match="Constant must be a number"):
        _ = ConstraintBinary(LINEAR_POLY, 0, 1) * PolyBinary()

def test_eq():
    """ test equality of constraints """
    # Equality Constraints: Upper_Bound == Lower_Bound
    constraint1 = ConstraintBinary(QUADRATIC_POLY, 2, 2)
    constraint2 = ConstraintBinary(QUADRATIC_POLY, 2, 2)
    assert constraint1 != QUADRATIC_POLY
    assert constraint1 == constraint2
    assert constraint1 == constraint2.normalize()
    assert constraint1.polynomial != constraint2.normalize().polynomial
    assert constraint1 == constraint1 + 3
    assert constraint1 == constraint1 * 4

    # Inequality Constraints: Upper_Bound > Lower_Bound
    constraint1 = ConstraintBinary(QUADRATIC_POLY, 5, 6)
    constraint2 = ConstraintBinary(PolyBinary({(("x", 1),): 7, (("x", 2), ("x", 1)): 2, (): -3}), 5, 6)
    assert constraint1 != QUADRATIC_POLY
    assert constraint1 == constraint2
    assert constraint1 == constraint2.normalize()
    assert constraint1.polynomial != constraint2.normalize().polynomial
    assert constraint1 == constraint1 + 3
    assert constraint1 == constraint1 * 4

    # different polynomial
    assert constraint1 != ConstraintBinary(PolyBinary({(("x", 1),): 5, (("x", 2), ("x", 1)): 2, (): 3}), 5, 6)
    # different upper bound
    assert constraint1 != ConstraintBinary(QUADRATIC_POLY, 5, 5)
    # different lower bound
    assert constraint1 != ConstraintBinary(QUADRATIC_POLY, 4, 6)
    # different variable types
    assert constraint1 != ConstraintBinary(QUADRATIC_POLY.compact(), 5, 6)
    # different monomials
    assert constraint1 != ConstraintBinary(PolyBinary({(("x", 2), ("x", 1)): 2, (): 5}), 5, 6)
    # empty polynomial
    with pytest.warns(UserWarning, match="Useless constraint '0 == 0'"):
        useless1 = ConstraintBinary(PolyBinary(), 0, 0)
        assert useless1 == ConstraintBinary(PolyBinary(), 0, 0)
    assert useless1 != constraint1

    with pytest.warns(UserWarning, match="Useless constraint '0 <= 0 <= 1'"):
        useless2 = ConstraintBinary(PolyBinary(), 0, 1)
    assert useless1 != useless2

    # equality constraints with str variables
    assert ConstraintBinary(PolyBinary({("a", "b"): 1}), 0, 0) == ConstraintBinary(PolyBinary({("a", "b"): 2}), 0, 0)

    # equality constraints with int variables
    constraint1 = ConstraintBinary(PolyBinary({(1, 2): 1, (3, 4): 1}), 1, 1)
    constraint2 = ConstraintBinary(PolyBinary({(1, 2): 3, (3, 4): 3, (): 2}), 5, 5)
    assert constraint1 == constraint2

def test_repr():
    """ test string representation """
    constraint = ConstraintBinary(PolyBinary({(0,): 1, (1,): 1}), 0, 1)
    assert repr(constraint) == "ConstraintBinary(PolyBinary({(0,): 1, (1,): 1}), 0, 1)"

def test_str():
    """ test string creation """
    constraint = ConstraintBinary(QUADRATIC_POLY, 1, 2)
    assert str(constraint) == "1 <= -3 +7 x_1 +2 x_1 x_2 <= 2"

    constraint = ConstraintBinary(QUADRATIC_POLY, 4, 4)
    assert str(constraint) == "-3 +7 x_1 +2 x_1 x_2 == 4"

    constraint = ConstraintBinary(QUADRATIC_POLY, 4, 4)
    assert constraint.__str__(revert_eq=True) == "4 == -3 +7 x_1 +2 x_1 x_2"  # pylint: disable=unnecessary-dunder-call

    constraint = ConstraintBinary(PolyBinary({(): -15.99999992050575, (('alpha', 0),): 1.0, (('alpha', 1),): 2.0,
                                              (('alpha', 2),): 4.5, (('x', 2),): -4.000000022959645, (('x', 3),):
                                                  -8.00000004591929, (('x', 4),): -1.0000000057399112}), -20.0, -9.0)
    assert str(constraint) == "-20 <= -16 +1 alpha_0 +2 alpha_1 +4.5 alpha_2 -4 x_2 -8 x_3 -1 x_4 <= -9"

def test_is_equality_constraint():
    """ test if constraint is equality """
    constraint = ConstraintBinary(LINEAR_POLY, 0, 1)
    assert not constraint.is_equality_constraint()
    constraint = ConstraintBinary(QUADRATIC_POLY, 0, 0)
    assert constraint.is_equality_constraint()
    constraint = ConstraintBinary(QUBIC_POLY, 1, 1)
    assert constraint.is_equality_constraint()

def test_is_integer():
    """ test if constraint is integer """
    constraint = ConstraintBinary(LINEAR_POLY, 0, 1)
    assert constraint.is_integer()
    constraint = ConstraintBinary(QUADRATIC_POLY, 0.5, 1)
    assert not constraint.is_integer()
    constraint = ConstraintBinary(QUBIC_POLY, 0, 0.5)
    assert not constraint.is_integer()

    poly = PolyBinary({(("x", 1), ("x", 2), ("x", 3)): 1.000001, (("x", 1),): 1.000001})
    constraint = ConstraintBinary(poly, 0, 1)
    assert not constraint.is_integer()
    assert constraint.is_integer(precision=1e-3)

def test_round():
    """ test rounding """
    poly = PolyBinary({(0,): 5, (1,): 6, (2, 0): -3, (1, 3): 0.5, (3,): -1, (0, 1): 0.9, (): 16, (2,): 0, (2, 2): 2})
    constraint = ConstraintBinary(poly, 12.4, 19.9)
    exp_poly = PolyBinary({(0,): 5, (1,): 6, (2, 0): -3, (2, 2): 2, (1, 3): 0, (3,): -1, (0, 1): 1, (): 16})
    exp_constraint = ConstraintBinary(exp_poly, 12, 20)
    assert constraint.round() == exp_constraint

    poly = PolyBinary({(1,): 4, (2,): 5.7777})
    constraint = ConstraintBinary(poly, 5, 8.991)
    exp_poly = PolyBinary({(1,): 4, (2,): 5.78})
    exp_constraint = ConstraintBinary(exp_poly, 5, 8.99)
    assert constraint.round(2) == exp_constraint

def test_is_normal():
    """ test check of normalization of constraints """
    assert ConstraintBinary(QUADRATIC_POLY, 0, 6).is_normal()
    assert not ConstraintBinary(QUADRATIC_POLY, 5, 6).is_normal()
    assert not ConstraintBinary(-1 * QUADRATIC_POLY, -6, 0).is_normal()

def test_normalize():
    """ test normalization of constraints """
    exp_constraint = ConstraintBinary(QUADRATIC_POLY, 0, 6)
    normalized = exp_constraint.normalize()
    assert normalized.polynomial == exp_constraint.polynomial
    assert normalized.lower_bound == exp_constraint.lower_bound
    assert normalized.upper_bound == exp_constraint.upper_bound

    constraint = ConstraintBinary(-1 * QUADRATIC_POLY, -6, 0)
    normalized = constraint.normalize()
    assert normalized.polynomial == exp_constraint.polynomial
    assert normalized.lower_bound == exp_constraint.lower_bound
    assert normalized.upper_bound == exp_constraint.upper_bound

    with pytest.warns(UserWarning, match="Useless constraint '0 <= 5 <= 10'"):
        constraint = ConstraintBinary(PolyBinary({(): 5}), 0, 10)
        assert constraint.normalize() == constraint

def test_replace_variables():
    """ test replacement of variables """
    constraint = ConstraintBinary(QUBIC_POLY, 0, 1)
    qubic_flat = PolyBinary({(0, 1, 2): 1, (0,): 1})
    exp_constraint = ConstraintBinary(qubic_flat, 0, 1)
    new_constraint = constraint.replace_variables({("x", 1): 0, ("x", 2): 1, ("x", 3): 2})
    assert new_constraint == exp_constraint

    replacement = {("x", 1): "c", ("x", 2): "b", ("x", 3): "a"}
    qubic_str = PolyBinary({("c", "b", "a"): 1, ("c",): 1})
    exp_constraint = ConstraintBinary(qubic_str, 0, 1)
    new_constraint = constraint.replace_variables(replacement)
    assert new_constraint == exp_constraint

def test_get_preprocessable_variables():
    """ test preprocessable variables """
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('x', 1\): 0, \('x', 2\): 0\}"):
        constraint = ConstraintBinary(LINEAR_POLY, 0, 0)
    assert constraint.get_preprocessable_variables() == {('x', 1): 0, ('x', 2): 0}

    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{'x': 1, 'y': 1\}"):
        constraint = ConstraintBinary(PolyBinary({("x",): 1, ("x", "y"): 2, (): 3}), 6, 6)
    assert constraint.get_preprocessable_variables() == {'x': 1, 'y': 1}

def test_preprocess():
    """ test preprocessable variables """
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('x', 1\): 0, \('x', 2\): 0\}"):
        constraint = ConstraintBinary(LINEAR_POLY, 0, 0)
    exp_preprocessed = {("x", 1): 0, ("x", 2): 0}
    preprocessed_constraint, preprocessed = constraint.preprocess()
    assert preprocessed_constraint is None
    assert preprocessed == exp_preprocessed

    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{'x': 1, 'y': 1\}"):
        constraint = ConstraintBinary(PolyBinary({("x",): 1, ("x", "y"): 2, ("z", "w"): -1, (): 3}), 6, 6)
    exp_constraint = ConstraintBinary(PolyBinary({("z", "w"): 1}), 0, 0)
    exp_preprocessed = {"x": 1, "y": 1}
    preprocessed_constraint, preprocessed = constraint.preprocess()
    assert preprocessed_constraint == exp_constraint
    assert preprocessed == exp_preprocessed

    with pytest.warns(UserWarning, match=r"The two variables 'x' and 'y' are set to be equal"):
        constraint = ConstraintBinary(PolyBinary({("x",): 1, ("y",): -1, (): 5}), 5, 5)
    exp_preprocessed = {"y": PolyBinary({"x": 1})}
    preprocessed_constraint, preprocessed = constraint.preprocess()
    assert preprocessed_constraint is None
    assert preprocessed == exp_preprocessed

    with pytest.warns(UserWarning, match=r"The two variables 'x' and 'y' are set to be negations of each other"):
        constraint = ConstraintBinary(PolyBinary({("x",): 1, ("y",): 1, (): 4}), 5, 5)
    exp_preprocessed = {"y": PolyBinary({"x": -1, (): 1})}
    preprocessed_constraint, preprocessed = constraint.preprocess()
    assert preprocessed_constraint is None
    assert preprocessed == exp_preprocessed

def test_copy():
    """ test copying of constraint """
    constraint = ConstraintBinary(LINEAR_POLY, 0, 1)
    constraint_copy = constraint.copy()
    assert constraint_copy == constraint
    assert constraint_copy is not constraint

def test_provides_penalty_term_directly():
    """ test check whether the polynomial of the constraint directly is a penalty term """
    constraint = ConstraintBinary(PolyBinary({(0, 1): 1, (1, 2): -1}), 0, 0)
    assert not constraint.provides_penalty_term_directly()

    # all summands positive
    constraint = ConstraintBinary(PolyBinary({(0, 1): 1, (1, 2): 1}), 0, 0)
    assert constraint.provides_penalty_term_directly()

    # penalty term to first linear reduction constraint
    constraint = ConstraintBinary(PolyBinary({('a', 'r'): 1, 'r': -1}), 0, 0)
    assert constraint.provides_penalty_term_directly()

    # penalty term to second linear reduction constraint
    constraint = ConstraintBinary(PolyBinary({('r',): 3, ('r', 'x'): -2, ('a', 'r'): -2, ('a', 'x'): 1}), 0, 0)
    assert constraint.provides_penalty_term_directly()

    # penalty term to second linear reduction constraint with different variable ordering
    constraint = ConstraintBinary(PolyBinary({('r',): 3, ('r', 'x'): -2, ('a', 'r'): -2, ('a', 'x'): 1}), 0, 0)
    assert not constraint.provides_penalty_term_directly(max_degree=1)

def test_get_reductions__standard():
    """ test reduction """
    # already linear, nothing happens
    constraint = ConstraintBinary(LINEAR_POLY, 3, 9)
    linear_constraint, reductions = constraint.get_reductions()
    assert linear_constraint == constraint
    assert not reductions

    with pytest.warns(UserWarning, match="Force is only necessary to get a fully linear problem"):
        constraint.get_reductions(max_degree=2, force=True)

    # already quadratic, nothing happens when reducing to degree 2
    constraint = ConstraintBinary(QUADRATIC_POLY, 3, 6)
    quadratic_constraint, reductions = constraint.get_reductions(2)
    assert quadratic_constraint == constraint
    assert not reductions

    # reduction from quadratic to linear
    linear_constraint, reductions = constraint.get_reductions()  # default reduction degree is 1
    exp_poly = PolyBinary({(("reduction", "x", 1, "x", 2),): 2, (("x", 1),): 7, (): -3})
    exp_constraint = ConstraintBinary(exp_poly, 3, 6)
    exp_reductions = [(('reduction', 'x', 1, 'x', 2), ('x', 1), ('x', 2))]
    assert linear_constraint == exp_constraint
    assert reductions == exp_reductions

    # reduction from qubic to quadratic
    constraint = ConstraintBinary(QUBIC_POLY, 0, 1)
    quadratic_constraint, reductions = constraint.get_reductions(2)
    exp_poly = PolyBinary({(('x', 1),): 1, (('reduction', 'x', 1, 'x', 2), ('x', 3)): 1})
    exp_constraint = ConstraintBinary(exp_poly, 0, 1)
    exp_reductions = [(('reduction', 'x', 1, 'x', 2), ('x', 1), ('x', 2))]
    assert quadratic_constraint == exp_constraint
    assert reductions == exp_reductions

    # reduction from qubic to linear
    linear_constraint, reductions = constraint.get_reductions(1)
    exp_poly = PolyBinary({(('x', 1),): 1, (('reduction', 'x', 3, 'reduction', 'x', 1, 'x', 2),): 1})
    exp_constraint = ConstraintBinary(exp_poly, 0, 1)
    exp_reductions = [(('reduction', 'x', 1, 'x', 2), ('x', 1), ('x', 2)),
                      (('reduction', 'x', 3, 'reduction', 'x', 1, 'x', 2), ('x', 3), ('reduction', 'x', 1, 'x', 2))]
    assert linear_constraint == exp_constraint
    assert reductions == exp_reductions

    # reduction from qubic to linear using a specific predefined reduction
    use_reductions = [(('nice_reduction_var', 1), ('x', 3), ('x', 1))]
    linear_constraint, reductions = constraint.get_reductions(1, use_reductions)
    exp_poly = PolyBinary({(('x', 1),): 1, (('reduction', 'nice_reduction_var', 1, 'x', 2),): 1})
    exp_constraint = ConstraintBinary(exp_poly, 0, 1)
    exp_reductions = [(('reduction', 'nice_reduction_var', 1, 'x', 2), ('nice_reduction_var', 1), ('x', 2))]
    assert linear_constraint == exp_constraint
    assert reductions == exp_reductions

    # just apply a specific predefined reduction
    use_reductions = [(('nice_reduction_var', 1), ('x', 3), ('x', 1))]
    reduced_constraint, reductions = constraint.get_reductions(None, use_reductions)
    exp_poly = PolyBinary({(('x', 1),): 1, (('nice_reduction_var', 1), ('x', 2)): 1})
    exp_constraint = ConstraintBinary(exp_poly, 0, 1)
    assert reduced_constraint == exp_constraint
    assert reductions == []

def test_get_reductions__special():
    """ test reduction on reduction constraint itself """
    # special constraint a * x == r
    constraint = ConstraintBinary(PolyBinary({('a', 'x'): -1, 'r': 1}), 0, 0)
    reduced_constraint, reductions = constraint.get_reductions()
    assert reduced_constraint is None
    assert reductions == [('r', 'a', 'x')]

    # special constraint 3 * r - 2 * r * x - 2 * a * r + a * x == 0  ->  reduction penalty term
    constraint = ConstraintBinary(PolyBinary({('r',): 3, ('r', 'x'): -2, ('a', 'r'): -2, ('a', 'x'): 1}), 0, 0)
    reduced_constraint, reductions = constraint.get_reductions(force=True)
    assert reduced_constraint is None
    assert reductions == [('r', 'a', 'x')]

    # special constraint 1 * r - 1 * r * x - 1 * a * r + a * x == 0  ->  reduction penalty term
    # but without force the constraint is kept
    constraint = ConstraintBinary(PolyBinary({('r',): 1, ('r', 'x'): -1, ('a', 'r'): -1, ('a', 'x'): 1}), 0, 0)
    reduced_constraint, reductions = constraint.get_reductions()
    assert reduced_constraint == constraint
    assert reductions == []

    reduced_constraint, reductions = constraint.get_reductions(force=True)
    assert reduced_constraint is None
    assert reductions == [('r', 'a', 'x')]

def test_reduce():
    """ test reduction and get all constraints """
    # reduction from quadratic to linear
    constraint = ConstraintBinary(QUADRATIC_POLY, 3, 6)
    linear_constraints = constraint.reduce(1)
    exp_poly = PolyBinary({(("reduction", "x", 1, "x", 2),): 2, (("x", 1),): 7, (): -3})
    exp_constraint_original = ConstraintBinary(exp_poly, 3, 6)
    exp_polys = [PolyBinary({(('x', 1),): -1, (('reduction', 'x', 1, 'x', 2),): 1}),
                 PolyBinary({(('x', 2),): -1, (('reduction', 'x', 1, 'x', 2),): 1}),
                 PolyBinary({(('x', 1),): -1, (('x', 2),): -1, (('reduction', 'x', 1, 'x', 2),): 1})]
    exp_constraints = {"original": exp_constraint_original,
                       "reduction_x_1_x_2_0": ConstraintBinary(exp_polys[0], -1, 0),
                       "reduction_x_1_x_2_1": ConstraintBinary(exp_polys[1], -1, 0),
                       "reduction_x_1_x_2_2": ConstraintBinary(exp_polys[2], -1, 0)}
    assert linear_constraints == exp_constraints

    # reduction from qubic to quadratic
    constraint = ConstraintBinary(QUBIC_POLY, 0, 1)
    quadratic_constraints = constraint.reduce(2)
    exp_poly = PolyBinary({(('x', 1),): 1, (('reduction', 'x', 1, 'x', 2), ('x', 3)): 1})
    exp_constraint_original = ConstraintBinary(exp_poly, 0, 1)
    exp_poly = PolyBinary({(('x', 1), ('x', 2)): -1, (('reduction', 'x', 1, 'x', 2),): 1})
    exp_constraint_reduction = ConstraintBinary(exp_poly, 0, 0)
    exp_constraints = {"original": exp_constraint_original,
                       "reduction_x_1_x_2": exp_constraint_reduction}
    assert quadratic_constraints == exp_constraints

    # just apply a specific predefined reduction
    use_reductions = [(("nice_reduction_var", 1), ('x', 3), ('x', 1))]
    reduced_constraints = constraint.reduce(None, use_reductions)
    exp_poly = PolyBinary({(('x', 1),): 1, (("nice_reduction_var", 1), ('x', 2)): 1})
    exp_constraint = ConstraintBinary(exp_poly, 0, 1)
    assert reduced_constraints == {"original": exp_constraint}

    # check useless constraint when replacing variables in useless constraint
    exp_constraints = {"reduction_x_1_x_2_0": ConstraintBinary(exp_polys[0], -1, 0),
                       "reduction_x_1_x_2_1": ConstraintBinary(exp_polys[1], -1, 0),
                       "reduction_x_1_x_2_2": ConstraintBinary(exp_polys[2], -1, 0)}
    with pytest.warns(UserWarning, match="Useless constraint"):
        constraint = ConstraintBinary(PolyBinary({(('x', 1), ('x', 2)): 1}), 0, 1)
        reduced_constraints = constraint.reduce(force=True, max_degree=1)
    assert reduced_constraints == exp_constraints


def test_get_penalty_terms__constant_polynomial():
    """ test construction of objective terms with constant polynomial """
    poly = PolyBinary({(): 5})
    with pytest.warns(UserWarning, match="Useless constraint '5 == 5'"):
        constraint = ConstraintBinary(poly, 5, 5)
        assert not constraint.get_penalty_terms()

    with pytest.warns(UserWarning, match="Useless constraint '0 <= 5 <= 10'"):
        constraint = ConstraintBinary(poly, 0, 10)
        assert not constraint.get_penalty_terms()

def test_get_penalty_terms__linear_equality_constraint():
    """ test construction of objective terms for equality constraint """
    constraint = ConstraintBinary(LINEAR_POLY, 5, 5)
    exp_penalty_terms = {NAME: (LINEAR_POLY - 5)**2}
    assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

    # in the following case the polynomial can be used itself as a penalty term as it is >=0 for all assignments
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('x', 1\): 0, \('x', 2\): 0\}"):
        constraint = ConstraintBinary(LINEAR_POLY, 0, 0)
    exp_penalty_terms = {NAME: LINEAR_POLY}
    assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

def test_get_penalty_terms__linear_inequality_constraint():
    """ test construction of objective terms for inequality """
    constraint = ConstraintBinary(LINEAR_POLY, 3, 9)
    exp_poly = (LINEAR_POLY - 3 - get_binary_representation_polynomial(6, variable_prefix="test_constraint_slack")) ** 2
    exp_penalty_terms = {NAME: exp_poly}
    assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

    with pytest.raises(ValueError, match="Inequality constraint must have integer coefficients and boundaries"):
        ConstraintBinary(PolyBinary({"x": 3.5}), 0, 1).get_penalty_terms()
    with pytest.raises(ValueError, match="Inequality constraint must have integer coefficients and boundaries"):
        ConstraintBinary(PolyBinary({"x": 3}), 0, 1.5).get_penalty_terms()

def test_get_penalty_terms__quadratic_equality_constraint():
    """ test construction of objective terms for equality constraint with quadratic polynomial """
    constraint = ConstraintBinary(QUADRATIC_POLY, 5, 5)
    poly, exp_reductions = QUADRATIC_POLY.reduce()
    exp_penalty_terms = {NAME: (poly - 5) ** 2}
    exp_penalty_terms.update(get_all_reduction_penalty_terms(exp_reductions))
    assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

    # in the following case the polynomial (minus the constant) can be used itself as a penalty term
    # as it is >=0 for all assignments
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('x', 1\): 0\}"):
        constraint = ConstraintBinary(QUADRATIC_POLY, -3, -3)
    exp_penalty_terms = {NAME: QUADRATIC_POLY + 3}
    assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

    constraint = ConstraintBinary(PolyBinary({("x",): 1, ("x", "y"): -2, (): 3}), 2, 2)
    exp_penalty_terms = {"reduction_x_y": PolyBinary({(("reduction", "x", "y"),): 3,
                                                      (("reduction", "x", "y"), ("x",)): -2,
                                                      (("reduction", "x", "y"), ("y",)): -2,
                                                      (("x",), ("y",)): 1}),
                         "test_constraint": PolyBinary({(): 1, (("x",),): 3, (("reduction", "x", "y"), ("x",)): -4})}
    with pytest.warns(UserWarning, match=r"Variables are replaced by tuple variables"):
        assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

def test_get_penalty_terms__quadratic_inequality_constraint():
    """ test construction of objective terms for inequality with quadratic polynomial """
    constraint = ConstraintBinary(QUADRATIC_POLY, 5, 6)
    poly, exp_reductions = QUADRATIC_POLY.reduce()
    exp_poly = (poly - 5 - get_binary_representation_polynomial(1, variable_prefix="test_constraint_slack")) ** 2
    exp_penalty_terms = {NAME: exp_poly}
    exp_penalty_terms.update(get_all_reduction_penalty_terms(exp_reductions))
    assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

def test_get_penalty_terms__quartic_equality_constraint():
    """ test construction of objective terms for inequality with quadratic polynomial """
    R, X = "reduction", "x"
    exp_reductions1 = {"reduction_x_1_x_2": PolyBinary({((R, X, 1, X, 2),): 3,
                                                        ((R, X, 1, X, 2), (X, 1)): -2,
                                                        ((R, X, 1, X, 2), (X, 2)): -2,
                                                        ((X, 1), (X, 2)): 1}),
                       "reduction_x_3_x_4": PolyBinary({((R, X, 3, X, 4),): 3,
                                                        ((R, X, 3, X, 4), (X, 3)): -2,
                                                        ((R, X, 3, X, 4), (X, 4)): -2,
                                                        ((X, 3), (X, 4)): 1})}
    exp_reductions2 = {"reduction_reduction_x_1_x_2_reduction_x_3_x_4":
                           PolyBinary({((R, R, X, 1, X, 2, R, X, 3, X, 4),): 3,
                                       ((R, R, X, 1, X, 2, R, X, 3, X, 4), (R, X, 1, X, 2)): -2,
                                       ((R, R, X, 1, X, 2, R, X, 3, X, 4), (R, X, 3, X, 4)): -2,
                                       ((R, X, 1, X, 2), (R, X, 3, X, 4)): 1})}

    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{1: 1, 2: 1, 3: 1, 4: 1\}"):
        constraint = ConstraintBinary(PolyBinary({(1, 2, 3, 4): 1}), 1, 1)

    exp_penalty_terms = {NAME: PolyBinary({(): 1, ((R, R, X, 1, X, 2, R, X, 3, X, 4),): -1})}
    exp_penalty_terms.update(exp_reductions1)
    exp_penalty_terms.update(exp_reductions2)
    with pytest.warns() as record:
        assert constraint.get_penalty_terms(NAME) == exp_penalty_terms
    assert len(record) == 2
    assert str(record[0].message) == "Variables are replaced by tuple variables"
    assert str(record[1].message) == "Variables could be set in advance: " \
                                     "{('reduction', 'reduction', 'x', 1, 'x', 2, 'reduction', 'x', 3, 'x', 4): 1}"

    constraint = ConstraintBinary(PolyBinary({((X, 1), (X, 2), (X, 3), (X, 4)): 1, ((X, 1), (X, 2)): 1}), 1, 1)
    exp_penalty_terms = {NAME: PolyBinary({(): 1,
                                           ((R, R, X, 1, X, 2, R, X, 3, X, 4),): -1,
                                           ((R, X, 1, X, 2),): -1,
                                           ((R, R, X, 1, X, 2, R, X, 3, X, 4), (R, X, 1, X, 2)): 2})}
    exp_penalty_terms.update(exp_reductions1)
    exp_penalty_terms.update(exp_reductions2)

    warning_msg = r"The two variables '\('reduction', 'reduction', 'x', 1, 'x', 2, 'reduction', 'x', 3, 'x', 4\)' " \
                  r"and '\('reduction', 'x', 1, 'x', 2\)' are set to be " \
                  r"negations of each other, consider replacing one with '1 - other'"
    with pytest.warns(UserWarning, match=warning_msg):
        assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

    constraint = ConstraintBinary(PolyBinary({((X, 1), (X, 2), (X, 3), (X, 4)): 1, ((X, 1), (X, 2)): 1}), 0, 0)
    exp_penalty_terms = {NAME: PolyBinary({((R, X, 1, X, 2),): 1, ((R, X, 1, X, 2), (R, X, 3, X, 4)): 1})}
    exp_penalty_terms.update(exp_reductions1)
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('reduction', 'x', 1, 'x', 2\): 0\}"):
        assert constraint.get_penalty_terms(NAME) == exp_penalty_terms

def test_get_penalty_terms__special_constraints():
    """ test construction of objective terms for special constraints """
    reduction_constraint1 = ConstraintBinary(PolyBinary({'x': 1, 'r': -1}), 0, 1)
    exp_penalty_terms = {'constraint': PolyBinary({('r',): 1, ('r', 'x'): -1})}
    assert reduction_constraint1.get_penalty_terms() == exp_penalty_terms

    reduction_constraint1 = ConstraintBinary(PolyBinary({'a': -1, 'r': 1, (): 1}), 0, 1)
    exp_penalty_terms = {'constraint': PolyBinary({('r',): 1, ('a', 'r'): -1})}
    assert reduction_constraint1.get_penalty_terms() == exp_penalty_terms

    reduction_constraint2 = ConstraintBinary(PolyBinary({'a': -1, 'r': 1, 'x': -1, (): 1}), 0, 1)
    exp_penalty_terms = {'reduction': PolyBinary({('r',): 1, ('a', 'r'): -1, ('r', 'x'): -1, ('a', 'x'): 1})}
    assert reduction_constraint2.get_penalty_terms('reduction') == exp_penalty_terms

    reduction_constraint0 = ConstraintBinary(PolyBinary({('a', 'x'): -1, 'r': 1, 's': 0}), 0, 0)
    exp_penalty_terms = {'constraint': PolyBinary({('r',): 3, ('a', 'r'): -2, ('r', 'x'): -2, ('a', 'x'): 1})}
    assert reduction_constraint0.get_penalty_terms() == exp_penalty_terms

def test_check_validity():
    """ test validity check """
    constraint = ConstraintBinary(PolyBinary({'x': 1, 'r': -1}), 0, 1)
    assert constraint.check_validity({'r': 0, 'x': 1})
    assert not constraint.check_validity({'r': 1, 'x': -1})

def test_get_special_constraint_penalty_poly():
    """ test check for reduction constraint """
    # linear constraint with 2 variables, 2 permutations
    reduction_constraint_l2 = ConstraintBinary(PolyBinary({'x': 1, 'r': -1}), 0, 1)
    exp_poly = PolyBinary({('r',): 1, ('r', 'x'): -1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l2) == exp_poly

    reduction_constraint_l2 = ConstraintBinary(PolyBinary({'a': -1, 'r': 1, (): 1}), 0, 1)
    exp_poly = PolyBinary({('r',): 1, ('a', 'r'): -1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l2) == exp_poly

    # penalty constraint with 2 variables, 2 permutations
    reduction_constraint_l2 = ConstraintBinary(PolyBinary({'r': 1, ('x', 'r'): -1}), 0, 0)
    exp_poly = PolyBinary({('r',): 1, ('r', 'x'): -1})
    poly = get_special_constraint_penalty_poly(reduction_constraint_l2)
    assert poly == exp_poly

    reduction_constraint_l2 = ConstraintBinary(PolyBinary({('a', 'r'): 1, 'r': -1}), 0, 0)
    exp_poly = PolyBinary({('r',): 1, ('a', 'r'): -1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l2) == exp_poly

    # linear constraint with 3 variables, 3 permutations, 1 shift
    reduction_constraint_l3 = ConstraintBinary(PolyBinary({'a': -1, 'r': 1, 'x': -1, (): 1}), 0, 1)
    exp_poly = PolyBinary({('r',): 1, ('a', 'r'): -1, ('r', 'x'): -1, ('a', 'x'): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l3) == exp_poly

    reduction_constraint_l3 = ConstraintBinary(PolyBinary({'x': 1, 'y': 1, 'r': -1}), 0, 1)
    exp_penalty_poly = PolyBinary({('r',): 1, ('r', 'x'): -1, ('r', 'y'): -1, ('x', 'y'): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l3) == exp_penalty_poly

    reduction_constraint_l3 = ConstraintBinary(PolyBinary({(1,): 1, (3,): 1, (2,): -1}), 0, 1)
    exp_penalty_poly = PolyBinary({(2,): 1, (2, 1): -1, (2, 3): -1, (1, 3): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l3) == exp_penalty_poly

    reduction_constraint_l3 = ConstraintBinary(PolyBinary({(('x', 1),): 1,
                                                           (('x', 2),): 1,
                                                           (('x', 3),): -1,
                                                           (): 3}), 3, 4)
    exp_penalty_poly = PolyBinary({(('x', 3),): 1,
                                   (('x', 3), ('x', 1)): -1,
                                   (('x', 3), ('x', 2)): -1,
                                   (('x', 1), ('x', 2)): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_l3) == exp_penalty_poly

    # penalty constraint with 3 variables, 1 shift
    penalty_constraint_l3 = ConstraintBinary(exp_penalty_poly + 5, 5, 5)
    assert get_special_constraint_penalty_poly(penalty_constraint_l3) == exp_penalty_poly

    # full penalty constraint
    exp_penalty_poly = PolyBinary({(('x', 3),): 3,
                                   (('x', 3), ('x', 1)): -2,
                                   (('x', 3), ('x', 2)): -2,
                                   (('x', 1), ('x', 2)): 1})
    penalty_constraint_full = ConstraintBinary(exp_penalty_poly, 0, 0)
    assert get_special_constraint_penalty_poly(penalty_constraint_full) == exp_penalty_poly

    # quadratic constraint, 3 permutations, 1 shift
    reduction_constraint_q = ConstraintBinary(PolyBinary({('a', 'x'): -1, 'r': 1}), 0, 0)
    exp_poly = PolyBinary({('r',): 3, ('a', 'r'): -2, ('r', 'x'): -2, ('a', 'x'): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_q) == exp_poly

    reduction_constraint_q = ConstraintBinary(PolyBinary({(0, 1): -1, (2,): 1, (): -3}), -3, -3)
    exp_poly = PolyBinary({(2,): 3, (0, 2): -2, (2, 1): -2, (0, 1): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_q) == exp_poly

    reduction_constraint_q = ConstraintBinary(PolyBinary({(('x', 2), ('x', 1)): -1, (('x', 0),): 1}), 0, 0)
    exp_poly = PolyBinary({(('x', 0),): 3,
                           (('x', 2), ('x', 0)): -2,
                           (('x', 0), ('x', 1)): -2,
                           (('x', 1), ('x', 2)): 1})
    assert get_special_constraint_penalty_poly(reduction_constraint_q) == exp_poly

def test_to_equality_constraint():
    """ test the construction of the corresponding equality constraint """
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('x', 1\): 0, \('x', 2\): 0\}"):
        constraint = ConstraintBinary(LINEAR_POLY, 0, 0)
    assert constraint.to_equality_constraint() == constraint

    constraint = ConstraintBinary(LINEAR_POLY, 0, 1)
    exp_poly = LINEAR_POLY - PolyBinary({(("slack", 0),) : 1})
    exp_constraint = ConstraintBinary(exp_poly, 0, 0)
    assert constraint.to_equality_constraint() == exp_constraint

    constraint = ConstraintBinary(QUADRATIC_POLY, 0, 3)
    exp_poly = QUADRATIC_POLY - PolyBinary({(("other_slack", 0),) : 1, (("other_slack", 1),) : 2})
    exp_constraint = ConstraintBinary(exp_poly, 0, 0)
    assert constraint.to_equality_constraint("other_slack") == exp_constraint

def test_get_standard_penalty_polynomial():
    """ test the construction of the penalty term for linear constraint """
    with pytest.warns(UserWarning, match=r"Variables could be set in advance: \{\('x', 1\): 0, \('x', 2\): 0\}"):
        constraint = ConstraintBinary(LINEAR_POLY, 0, 0)
    exp_poly = LINEAR_POLY ** 2
    assert get_standard_penalty_polynomial(constraint) == exp_poly

    constraint = ConstraintBinary(LINEAR_POLY, 0, 1)
    exp_poly = (LINEAR_POLY - PolyBinary({(("slack", 0),) : 1})) ** 2
    assert get_standard_penalty_polynomial(constraint) == exp_poly

    constraint = ConstraintBinary(QUADRATIC_POLY, 0, 1)
    exp_poly = PolyBinary({(): 9, (('slack', 0),): 7, (('x', 1),): 7, (('slack', 0), ('x', 1)): -14,
                           (('x', 1), ('x', 2)): 20, (('slack', 0), ('x', 1), ('x', 2)): -4})
    assert get_standard_penalty_polynomial(constraint) == exp_poly

def test_get_reduction_constraints():
    """ check the reduction constraints """
    exp_constraint1 = ConstraintBinary(PolyBinary({(('r', 2, 5),): -1, (('x', 2),): 1}), 0, 1)
    exp_constraint2 = ConstraintBinary(PolyBinary({(('r', 2, 5),): -1, (('x', 5),): 1}), 0, 1)
    exp_constraint3 = ConstraintBinary(PolyBinary({(('r', 2, 5),): -1, (('x', 2),): 1, (('x', 5),): 1}), 0, 1)
    exp_constraints = {"r_2_5_0": exp_constraint1, "r_2_5_1": exp_constraint2, "r_2_5_2": exp_constraint3}
    assert get_reduction_constraints(("r", 2, 5), ("x", 2), ("x", 5)) == exp_constraints

    exp_constraint = ConstraintBinary(PolyBinary({("r_2_5",): 1, ("x_2", "x_5"): -1}), 0, 0)
    exp_constraints = {"r_2_5": exp_constraint}
    assert get_reduction_constraints("r_2_5", "x_2", "x_5", max_degree=2) == exp_constraints

def test_get_all_reduction_penalty_terms():
    """ test the penalty terms from the reductions """
    reductions = [(('reduction', 'x', 1, 'x', 2), ('x', 1), ('x', 2)),
                  (('reduction', 'x', 2, 'x', 3), ('x', 2), ('x', 3)),
                  (('reduction', 'x', 3, 'reduction', 'x', 1, 'x', 2), ('x', 3), ('reduction', 'x', 1, 'x', 2)),
                  (('reduction', 'x', 4, 'reduction', 'x', 2, 'x', 3), ('x', 4), ('reduction', 'x', 2, 'x', 3))]
    exp_penalty_term1 = PolyBinary({(("reduction", "x", 1, "x", 2),): 3,
                                    (("reduction", "x", 1, "x", 2), ("x", 1)): -2,
                                    (("reduction", "x", 1, "x", 2), ("x", 2)): -2,
                                    (("x", 1), ("x", 2)): 1})
    exp_penalty_term2 = PolyBinary({(("reduction", "x", 2, "x", 3),): 3,
                                    (("reduction", "x", 2, "x", 3), ("x", 2)): -2,
                                    (("reduction", "x", 2, "x", 3), ("x", 3)): -2,
                                    (("x", 2), ("x", 3)): 1})
    exp_penalty_term3 = PolyBinary({(("reduction", "x", 3, "reduction", "x", 1, "x", 2),): 3,
                                    (("reduction", "x", 1, "x", 2),
                                     ("reduction", "x", 3, "reduction", "x", 1, "x", 2)): -2,
                                    (("reduction", "x", 1, "x", 2), ("x", 3)): 1,
                                    (("reduction", "x", 3, "reduction", "x", 1, "x", 2), ("x", 3)): -2})
    exp_penalty_term4 = PolyBinary({(("reduction", "x", 4, "reduction", "x", 2, "x", 3),): 3,
                                    (("reduction", "x", 2, "x", 3),
                                     ("reduction", "x", 4, "reduction", "x", 2, "x", 3)): -2,
                                    (("reduction", "x", 2, "x", 3), ("x", 4)): 1,
                                    (("reduction", "x", 4, "reduction", "x", 2, "x", 3), ("x", 4)): -2})
    exp_penalty_terms = {"reduction_x_1_x_2": exp_penalty_term1,
                         "reduction_x_2_x_3": exp_penalty_term2,
                         "reduction_x_3_reduction_x_1_x_2": exp_penalty_term3,
                         "reduction_x_4_reduction_x_2_x_3": exp_penalty_term4}
    assert get_all_reduction_penalty_terms(reductions) == exp_penalty_terms


## exhaustive tests for special constraints

VARS_STANDARD = ("x", 1), ("x", 2), ("reduction", "x", 1, "x", 2)
VARS_RESORTED_1 = ("a", 1), ("b", 2), ("reduction", "a", 1, "b", 2)
VARS_RESORTED_2 = ("a", 1), ("y", 2), ("reduction", "a", 1, "y", 2)


@pytest.fixture(params=[VARS_STANDARD, VARS_RESORTED_1, VARS_RESORTED_2],
                ids=["standard", "resorted_1", "resorted_2"],
                name="reduction_constraints")
def fixture_reduction_constraints(request):
    """ get test objects """
    # pylint: disable=too-many-locals
    x1, x2, r = request.param
    constraint_poly_lin2_1 = PolyBinary({(r,): 1, (x1,): -1, (): 1})
    constraint_poly_lin2_2 = PolyBinary({(r,): 1, (x2,): -1, (): 1})
    constraint_poly_lin3   = PolyBinary({(r,): 1, (x1,): -1, (x2,): -1, (): 1})
    constraint_poly_quad   = PolyBinary({(r,): 1, (x1, x2): -1})
    penalty_poly_lin2_1    = PolyBinary({(r,): 1, (r, x1): -1})
    penalty_poly_lin2_2    = PolyBinary({(r,): 1, (r, x2): -1})
    penalty_poly_lin3      = PolyBinary({(r,): 1, (r, x1): -1, (r, x2): -1, (x1, x2): 1})
    penalty_poly_full      = PolyBinary({(r,): 3, (r, x1): -2, (r, x2): -2, (x1, x2): 1})

    constraint_lin2_1         = ConstraintBinary(constraint_poly_lin2_1, 0, 1)
    constraint_lin2_2         = ConstraintBinary(constraint_poly_lin2_2, 0, 1)
    constraint_lin3           = ConstraintBinary(constraint_poly_lin3, 0, 1)
    constraint_quad           = ConstraintBinary(constraint_poly_quad, 0, 0)
    constraint_penalty_lin2_1 = ConstraintBinary(penalty_poly_lin2_1, 0, 0)
    constraint_penalty_lin2_2 = ConstraintBinary(penalty_poly_lin2_2, 0, 0)
    constraint_penalty_lin3   = ConstraintBinary(penalty_poly_lin3, 0, 0)
    constraint_penalty_full   = ConstraintBinary(penalty_poly_full, 0, 0)

    all_combinations =  [(constraint_lin2_1,         penalty_poly_lin2_1, False, (r, x1)),
                         (constraint_lin2_2,         penalty_poly_lin2_2, False, (r, x2)),
                         (constraint_lin3,           penalty_poly_lin3,   False, (r, x1, x2)),
                         (constraint_quad,           penalty_poly_full,   False, (r, x1, x2)),
                         (constraint_penalty_lin2_1, penalty_poly_lin2_1, True,  (r, x1)),
                         (constraint_penalty_lin2_2, penalty_poly_lin2_2, True,  (r, x2)),
                         (constraint_penalty_lin3,   penalty_poly_lin3,   True,  (r, x1, x2)),
                         (constraint_penalty_full,   penalty_poly_full,   True,  (r, x1, x2))]
    yield all_combinations

@pytest.fixture(params=range(8),
                ids=["lin2_1", "lin2_2", "lin3", "quad", "p_lin2_2", "p_lin2_2", "p_lin3", "p_full"],
                name="reduction_constraint__penalty_poly__is_penalty__reduction")
def fixture_reduction_constraint__penalty_poly__is_penalty__reduction(request, reduction_constraints):
    """ address each test object individually """
    yield reduction_constraints[request.param]


def test_provides_penalty_term_directly__reduction(reduction_constraint__penalty_poly__is_penalty__reduction):
    """ test check whether the polynomial of the constraint directly is a penalty term """
    constraint, _, provides_penalty_term, _ = reduction_constraint__penalty_poly__is_penalty__reduction
    if provides_penalty_term:
        assert constraint.provides_penalty_term_directly()
    else:
        assert not constraint.provides_penalty_term_directly()
    assert not constraint.provides_penalty_term_directly(max_degree=1)

def test_retrieve_reductions(reduction_constraint__penalty_poly__is_penalty__reduction):
    """ test retrieval of reductions from the constraint """
    constraint, _, _, exp_reduction = reduction_constraint__penalty_poly__is_penalty__reduction
    reductions = retrieve_reduction(constraint)
    assert reductions == exp_reduction

def test_get_reductions(reduction_constraint__penalty_poly__is_penalty__reduction):
    """ test getting the reductions from the constraint """
    constraint, _, provides_penalty_poly, exp_reduction = reduction_constraint__penalty_poly__is_penalty__reduction
    reduced_constraint, reductions = constraint.get_reductions(force=True)
    assert reduced_constraint is None
    assert reductions == [exp_reduction]

    exp_reductions_non_force = [] if provides_penalty_poly or constraint.polynomial.degree <= 1 else [exp_reduction]
    exp_constraint = constraint if provides_penalty_poly or constraint.polynomial.degree <= 1 else None
    reduced_constraint, reductions = constraint.get_reductions()
    assert reduced_constraint == exp_constraint
    assert reductions == exp_reductions_non_force

def test_matches_special_penalty_constraint__reduction(reduction_constraint__penalty_poly__is_penalty__reduction):
    """ test recognition of special constraints """
    constraint, _, provides_penalty_poly, _ = reduction_constraint__penalty_poly__is_penalty__reduction
    assert (provides_penalty_poly and matches_special_penalty_constraint(constraint)) \
           or (not provides_penalty_poly and not matches_special_penalty_constraint(constraint))

def test_get_special_constraint_penalty_poly__reduction(reduction_constraint__penalty_poly__is_penalty__reduction):
    """ test penalty polynomial of special constraint """
    constraint, exp_penalty_poly, _, _ = reduction_constraint__penalty_poly__is_penalty__reduction
    penalty_poly = get_special_constraint_penalty_poly(constraint)
    assert penalty_poly == exp_penalty_poly



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for HardwareAdjacency """

from functools import cached_property
import networkx as nx


NODE_DATA = "node_data"
EDGE_DATA = "edge_data"


class HardwareAdjacency(list):
    """
    A hardware adjacency is a list of edges, pairs of nodes, which implicitly defines the hardware graph.
    """

    def __init__(self, edges, name=None):
        """
        initialize HardwareAdjacency object

        :param (set or list) edges: set of tuples representing edges in the hardware graph
        :param (str or None) name: identifying name to differ between several hardware adjacencies
        """
        self.name = name
        super().__init__(_sort_edges(edges))

    def __eq__(self, other):
        # we also allow for the comparison with plain lists where the name is irrelevant
        if isinstance(other, list) and super().__eq__(_sort_edges(other)):
            if isinstance(other, HardwareAdjacency):
                return self.name == other.name
            return True
        return False

    def __ne__(self, other):
        # needs to be implemented otherwise would just use list comparison
        return not self.__eq__(other)

    def __repr__(self):
        """ get nice string representation """
        return self.__class__.__name__ + f"({super().__repr__()}, {repr(self.name)})"

    def __str__(self):
        """ get nice human-readable string representation of the objective """
        return str(list(self))

    @cached_property
    def nodes(self):
        """ nodes of the hardware graph """
        return sorted(set(node for edge in self for node in edge))

    @cached_property
    def graph(self):
        """ the corresponding networkx graph """
        graph = nx.Graph()
        graph.add_nodes_from(self.nodes)
        graph.add_edges_from(self)
        return graph

    def are_neighbored(self, node1, node2):
        """
        check whether two nodes are connected by an edge

        :param node1: first node in hardware graph
        :param node2: second node in hardware graph
        :return: True if the nodes are neighbored
        """
        return tuple(sorted((node1, node2))) in self

    def update_graph(self, node_data=None, edge_data=None):
        """
        update the graph's nodes or edges with the given data

        :param (dict or None) node_data: dictionary with additional information for the nodes,
                                         should have format {info_names: single_value_for_all_nodes or {nodes: values}}
        :param (dict or None) edge_data: dictionary with additional information for the edges,
                                         should have format {info_names: single_value_for_all_edges or {edges: values}}
        """
        if node_data:
            nx.set_node_attributes(self.graph, _revert_data(self.graph.nodes, node_data))
            self.graph.graph[NODE_DATA] = self.graph.graph.get(NODE_DATA, set()).union(node_data.keys())
        if edge_data:
            nx.set_edge_attributes(self.graph, _revert_data(self.graph.edges, edge_data))
            self.graph.graph[EDGE_DATA] = self.graph.graph.get(EDGE_DATA, set()).union(edge_data.keys())


def _sort_edges(edges):
    """ sort the elements in the tuples and afterwards the tuples """
    return sorted(set(tuple(sorted(edge)) for edge in edges))

def _revert_data(elements, data):
    """ extract the data for each element """
    return {e: {name: d if not isinstance(d, dict) else d[e] for name, d in data.items()} for e in elements}



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the helper functions for storing different types of Polynomials """

import os
import pytest
import h5py

from quark import PolyBinary, PolyIsing
from quark.io import hdf5_attributes
from quark.io.poly_utils import write_poly_with_type, read_poly_with_type


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE = "poly_utils.h5"
FILENAME_TMP = TEST_DIR + FILE
FILENAME_FIX = TEST_DIR + "testdata/" + FILE
GROUP_NAME = "test_group"

NAME = "test_poly_ising"
POLY_ISING = PolyIsing({(("y", 2, 3), ("x", 1)): 2, (("y", 1, 2),): 3}, inverted=True)
POLY_BINARY = PolyBinary({(("x", 1),): 2.5, (("x", 2), ("y", 0, 0)): 2.0, (): 5.5})


def test_write_poly_with_type():
    with h5py.File(FILENAME_TMP, "w") as h5file:
        group = h5file.create_group(GROUP_NAME)

        write_poly_with_type(group, POLY_ISING, NAME)

        assert NAME in group
        poly_group = group[NAME]
        assert "type" in poly_group.attrs
        type_str = hdf5_attributes.read_attribute(poly_group, "type")
        assert type_str == "PolyIsing"

    os.remove(FILENAME_TMP)

def test_read_poly_with_type():
    with h5py.File(FILENAME_FIX, "r") as h5file:
        group = h5file[GROUP_NAME]

        # load Polynomial stored under a name
        poly = read_poly_with_type(group, NAME)
        assert isinstance(poly, PolyIsing)
        assert poly == POLY_ISING

        message = f"Did not find Polynomial 'wrong_name' in group '/test_group' in hdf5 file '.*{FILE}'"
        with pytest.raises(ValueError, match=message):
            read_poly_with_type(group, "wrong_name")

        message = "Found too much Polynomials in group '/test_group'"
        with pytest.raises(ValueError, match=message):
            # there are too much unnamed Polynomials in this group
            read_poly_with_type(group)

        # but if the name corresponds to a valid type we can load it nevertheless
        poly = read_poly_with_type(group, name="PolyBinary")
        assert isinstance(poly, PolyBinary)
        assert poly == POLY_BINARY

        # if the unnamed Polynomial is the only one in the group we can load it, too
        # -> will search in group for PolyIsing, PolyBinary or Polynomial
        group = h5file["without_name"]
        poly = read_poly_with_type(group)
        assert isinstance(poly, PolyBinary)
        assert poly == POLY_BINARY

        group = h5file["empty_group"]
        message = "Found no Polynomials in group '/empty_group'"
        with pytest.raises(ValueError, match=message):
            # there no Polynomials in this group
            read_poly_with_type(group)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the IO of the PolyIsing """

import os
import pytest

from quark import PolyIsing
from quark.io import hdf5, txt


TEST_DIR = os.path.dirname(os.path.abspath(__file__)) + "/../"
FILE_H5 = "poly_ising.h5"
FILE_TXT = "poly_ising.txt"
FILENAME_H5_TMP = TEST_DIR + FILE_H5
FILENAME_TXT_TMP = TEST_DIR + FILE_TXT
FILENAME_H5_FIX = TEST_DIR + "testdata/" + FILE_H5
FILENAME_TXT_FIX = TEST_DIR + "testdata/" + FILE_TXT

POLY_FIX = PolyIsing({(("x", 1),): 2.5, (("x", 2), ("y", 0, 0)): 2.0, (): 5.5})
POLYS = [PolyIsing({(1,): 5, (2,): 6, (3, 1): 3, (2, 4): 3, (4,): 1, (): 3}),
         PolyIsing({(2,): 1, (2, 4): 1, (3, 4): 3, (5,): 7}, inverted=True),
         PolyIsing({(1,): 7, (2,): 2, (3,): 3, (4,): 1}),
         PolyIsing({}, inverted=True),
         PolyIsing({(): 55}),
         PolyIsing({(): 55, (1,): 66}, inverted=True),
         PolyIsing({(): 55, (1,): 66, (1, 2): 77}),
         PolyIsing({(1,): 66, (1, 2): 77}, inverted=True),
         PolyIsing({(1,): 1.5, (2,): 2.5}),
         PolyIsing({((1,),): 1.0, ((1, 2),): 2.0, ((1, 2), (1, 2), (1, 3)): 2.0, (): 5.5}, inverted=True),
         POLY_FIX]


def test_load():
    """ test hdf5 io """
    loaded = hdf5.load(PolyIsing, FILENAME_H5_FIX)
    assert loaded == POLY_FIX

    loaded = txt.load(PolyIsing, FILENAME_TXT_FIX)
    assert loaded == POLY_FIX

@pytest.mark.parametrize("poly", POLYS)
def test_io(poly):
    """ test hdf5 io round trip """
    hdf5.save(poly, FILENAME_H5_TMP)
    loaded = hdf5.load(PolyIsing, FILENAME_H5_TMP)
    assert loaded == poly
    assert loaded.is_inverted() == poly.is_inverted()
    os.remove(FILENAME_H5_TMP)

    txt.save(poly, FILENAME_TXT_TMP)
    loaded = txt.load(PolyIsing, FILENAME_TXT_TMP)
    assert loaded == poly
    assert loaded.is_inverted() == poly.is_inverted()
    os.remove(FILENAME_TXT_TMP)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for Instance """

from abc import ABCMeta, abstractmethod

from . import hdf5


class NotSupportedError(NotImplementedError):
    """ additional error to indicate that the functionality is not supported (and is therefore not implemented) """

    def __init__(self, message="This functionality is not supported"):
        """ set a default message """
        super().__init__(message)


class Instance:
    """
    An instance is a container for all data defining a specific instance of a problem.
    This class serves as a base class with methods for saving in and loading from HDF5 files.
    It only needs to be inherited if IO functionality is needed.
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        """ initialize Instance object and check the consistency of the data """
        self.check_consistency()

    @abstractmethod
    def check_consistency(self):
        """
        check if the data is consistent,
        needs to be implemented in inheriting class
        """
        raise NotImplementedError()

    @abstractmethod
    def get_name(self):
        """
        get an expressive name representing the instance,
        needs to be implemented in inheriting class

        :return: an expressive name representing the instance
        """
        raise NotImplementedError()

    def get_constrained_objective(self, name=None):
        """
        get the constrained objective object based on this instance data,

        implement this method in your inheriting instance class to link to implemented constrained objective
        (usually you have implemented either ConstrainedObjective, ObjectiveTerms or Objective),
        do this by using the method `get_from_instance(self)`, which is inherited from the base ConstrainedObjective

        :param (str or None) name: the name of the object
        :return: the instantiated constrained objective
        """
        raise NotSupportedError()

    def get_objective_terms(self, *args, **kwargs):
        """
        get the objective terms object based on this instance data,

        this defaults back to the implementation of the constrained objective,

        overwrite this method in your inheriting instance class to link to the objective terms, if you have implemented
        this instead (usually you have implemented either ConstrainedObjective, ObjectiveTerms or Objective),
        do this by using the method `get_from_instance(self)`, which is inherited from the base ObjectiveTerms

        :return: the instantiated objective terms
        """
        constrained_objective = self.get_constrained_objective()
        return constrained_objective.get_objective_terms(*args, **kwargs)

    def get_objective(self, *args, **kwargs):
        """
        get the objective object based on this instance data,

        this defaults back to the implementation of the objective terms,

        overwrite this method in your inheriting instance class to link to the objective, if you have implemented
        this instead (usually you have implemented either ConstrainedObjective, ObjectiveTerms or Objective),
        do this by using the method `get_from_instance(self)`, which is inherited from the base Objective

        :return: the instantiated objective
        """
        objective_terms = self.get_objective_terms()
        return objective_terms.get_objective(*args, **kwargs)

    @abstractmethod
    def write_hdf5(self, group):
        """
        write data in attributes and datasets in the opened group of the HDF5 file,
        needs to be implemented in inheriting class, is called by 'save_hdf5'

        :param (h5py.Group) group: the group to store the data in
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def read_hdf5(group):
        """
        read data from attributes and datasets of the opened group of the HDF5 file,
        needs to be implemented in inheriting class, is called by 'load_hdf5'

        :param (h5py.Group) group: the group to read the data from
        :return: the data as keyword arguments
        """
        raise NotImplementedError()

    @staticmethod
    def get_identifying_attributes():
        """
        get the attributes that identify an object of the class,
        eventually needs to be overwritten in inheriting class
        """
        return []

    def save_hdf5(self, filename, prefix_group_name=None, full_group_name=None, mode='a'):
        """ save instance in an HDF5 file """
        hdf5.save(self, filename, prefix_group_name, full_group_name, mode)

    @classmethod
    def load_hdf5(cls, filename, prefix_group_name=None, full_group_name=None, **identifiers):
        """ load instance from an HDF5 file """
        return hdf5.load(cls, filename, prefix_group_name, full_group_name, **identifiers)

    @classmethod
    def exists_hdf5(cls, filename, prefix_group_name=None, full_group_name=None, **identifiers):
        """ check if an instance already exists in an HDF5 file """
        return hdf5.exists(cls, filename, prefix_group_name, full_group_name, **identifiers)



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for VariableMapping """

from functools import cached_property
from bidict import bidict

from .utils.variables import get_common_type, check_type_against, to_string


KEY   = "key"
VALUE = "value"
TYPE  = "type_{}s"

ERROR_INCONSISTENT = "The {}s have inconsistent types"
ERROR_INVALID =      "Invalid {} type"


class VariableMapping(bidict):
    """ A VariableMapping stores the variable conversion of Polynomials. """

    def __init__(self, variables=None):
        """
        initialize VariableMapping object

        :param (list or dict) variables: the mapping of the variables,
                                         if it is a list, the indices are mapped to the variables
        """
        if isinstance(variables, list):
            variables = dict(enumerate(variables))
        elif variables is None:
            variables = {}  # or raise a specific error if None is not allowed
        super().__init__(variables)

    @cached_property
    def type_keys(self):
        """ cache for the type of the keys """
        try:
            return get_common_type(list(self.keys()))
        except TypeError as te:
            raise ValueError(ERROR_INCONSISTENT.format(KEY)) from te

    @cached_property
    def type_values(self):
        """ cache for the type of the values """
        try:
            return get_common_type(list(self.values()))
        except TypeError as te:
            raise ValueError(ERROR_INCONSISTENT.format(VALUE)) from te

    def _clear_types(self):
        """ clear the cached types """
        for attr in [TYPE.format(KEY), TYPE.format(VALUE)]:
            for obj in [self, self.inverse]:
                try:
                    delattr(obj, attr)
                except AttributeError:
                    pass

    def _write(self, newkey, newval, oldkey, oldval, unwrites):
        """ extends 'bidict._prep_write' to check for the cached types """
        try:
            check_type_against(newkey, self.type_keys)
        except TypeError as te:
            raise ValueError(ERROR_INVALID.format(KEY)) from te
        try:
            check_type_against(newval, self.type_values)
        except TypeError as te:
            raise ValueError(ERROR_INVALID.format(VALUE)) from te
        if not self:
            self._clear_types()
        return super()._write(newkey, newval, oldkey, oldval, unwrites)

    def _pop(self, key):
        """ extends 'bidict._pop' to clear cached types if necessary """
        v = super()._pop(key)
        if not self:
            self._clear_types()
        return v

    def clear(self):
        """
        remove all elements from the dictionary,
        extends 'bidict.clear' to clear cached types
        """
        super().clear()
        self._clear_types()

    def __str__(self):
        """ get nice human-readable string representation of the variable mapping """
        sorted_vars = sorted(self.keys())
        max_length = max(len(to_string(var)) for var in sorted_vars)
        return "\n".join(to_string(var) + " " * (max_length - len(to_string(var))) + f" <-> {to_string(self[var])}"
                         for var in sorted_vars)



# Copyright 2022 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing the HardwareAdjacency """

import pytest

from quark import HardwareAdjacency


@pytest.fixture(name="hwa")
def fixture_hwa():
    """ provide the HWA test object """
    yield HardwareAdjacency([(0, 1), (1, 2), (2, 0), (1, 3)], "test_hwa")


def test_init(hwa):
    """ test initialization of hwa """
    assert hwa == [(0, 1), (0, 2), (1, 2), (1, 3)]
    assert hwa.nodes == [0, 1, 2, 3]

    hwa2 = HardwareAdjacency([(0, 1), (1, 2), (2, 0), (1, 3)], "test_hwa")
    assert hwa2 == hwa
    hwa2 = HardwareAdjacency([(0, 1), (1, 2)], "test_hwa")
    assert hwa2 != hwa
    assert hwa2 == [(0, 1), (1, 2)]
    assert hwa2 == [(0, 1), (2, 1)]
    assert hwa2 != [(0, 1), (1, 3)]
    assert hwa2 != {(0, 1), (1, 2)}

def test_repr(hwa):
    """ test string representation """
    exp_str = "HardwareAdjacency([(0, 1), (0, 2), (1, 2), (1, 3)], 'test_hwa')"
    assert repr(hwa) == exp_str

def test_str(hwa):
    """ test string representation """
    exp_str = "[(0, 1), (0, 2), (1, 2), (1, 3)]"
    assert str(hwa) == exp_str

def test_are_neighbored(hwa):
    """ check neighborhood check """
    assert hwa.are_neighbored(0, 1)
    assert hwa.are_neighbored(1, 0)
    assert not hwa.are_neighbored(0, 3)
    assert not hwa.are_neighbored(3, 0)

def test_get_graph(hwa):
    """ test graph construction """
    node_data = {"pos": {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1)}}
    edge_data = {"color": 0}
    hwa.update_graph(node_data, edge_data)
    assert sorted(hwa.graph.nodes) == [0, 1, 2, 3]
    assert sorted(hwa.graph.edges) == [(0, 1), (0, 2), (1, 2), (1, 3)]
    assert dict(hwa.graph.nodes(data="pos")) == {0: (0, 0), 1: (1, 0), 2: (0, 1), 3: (1, 1)}
    assert list(hwa.graph.edges(data="color")) == [(0, 1, 0), (0, 2, 0), (1, 2, 0), (1, 3, 0)]



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for testing variable mapping """

from numbers import Integral
import pytest
from numpy.random import randint
from bidict import inverted, KeyAndValueDuplicationError, ValueDuplicationError

from quark import VariableMapping, Polynomial


X, Y = "x", "y"

MAPPING_DICT = {(1, 2): 4, (1, 3): 5}
MAPPING_DICT_INV = {4: (1, 2), 5: (1, 3)}


def test_init():
    """ test correct construction with inverse """
    # pylint: disable=unsubscriptable-object
    var_mapping = VariableMapping(MAPPING_DICT)
    assert isinstance(var_mapping.inverse, VariableMapping)
    assert var_mapping == MAPPING_DICT
    assert var_mapping.inverse == MAPPING_DICT_INV
    assert var_mapping[(1, 2)] == 4
    assert var_mapping[(1, 3)] == 5
    assert var_mapping.inverse[4] == (1, 2)
    assert var_mapping.inverse[5] == (1, 3)

    var_mapping = VariableMapping(inverted(MAPPING_DICT))
    assert isinstance(var_mapping.inverse, VariableMapping)
    assert var_mapping.inverse == MAPPING_DICT
    assert var_mapping == MAPPING_DICT_INV
    assert var_mapping.inverse[(1, 2)] == 4
    assert var_mapping.inverse[(1, 3)] == 5
    assert var_mapping[4] == (1, 2)
    assert var_mapping[5] == (1, 3)

    var_mapping = VariableMapping()
    assert isinstance(var_mapping.inverse, VariableMapping)
    assert not var_mapping
    assert not var_mapping.inverse
    assert "type_values" not in var_mapping.__dict__
    assert "type_keys" not in var_mapping.__dict__
    assert "type_values" not in var_mapping.inverse.__dict__
    assert "type_keys" not in var_mapping.inverse.__dict__
    assert var_mapping.type_values is None
    assert var_mapping.type_keys is None
    assert var_mapping.inverse.type_values is None
    assert var_mapping.inverse.type_keys is None

    with pytest.raises(ValueDuplicationError):
        VariableMapping({(1, 2): 4, (1, 3): 4})

def test_wrong_init():
    """ test correct raising of errors when instantiating with wrong variable types """
    with pytest.raises(ValueError, match="Invalid key type"):
        VariableMapping({X : 4, (1, 3) : 5})
    with pytest.raises(ValueError, match="Invalid value type"):
        VariableMapping({4 : X, 5 : (1, 3)})
    with pytest.raises(ValueError, match="Invalid value type"):
        VariableMapping({4 : True, 5 : False})
    with pytest.raises(ValueError, match="Invalid key type"):
        VariableMapping({True : 4, False : 5})

    var_mapping = VariableMapping()
    var_mapping.__dict__["_fwdm"] = {(X, 1) : 1, "x2" : 2}
    with pytest.raises(ValueError, match="The keys have inconsistent types"):
        _ = var_mapping.type_keys

    var_mapping = VariableMapping()
    var_mapping.__dict__["_invm"] = {(X, 1) : 1, "x2" : 2}
    with pytest.raises(ValueError, match="The values have inconsistent types"):
        _ = var_mapping.type_values

def test_set_item():
    """ test item setting """
    # pylint: disable=unsupported-assignment-operation
    var_mapping = VariableMapping(MAPPING_DICT)

    var_mapping[(2, 3)] = 3
    assert var_mapping == {(2, 3): 3, (1, 2): 4, (1, 3): 5}
    assert var_mapping.inverse == {3: (2, 3), 4: (1, 2), 5: (1, 3)}

    var_mapping[(2, 3)] = 6
    assert var_mapping == {(2, 3): 6, (1, 2): 4, (1, 3): 5}
    assert var_mapping.inverse == {6: (2, 3), 4: (1, 2), 5: (1, 3)}

    var_mapping.inverse[7] = (0, 1)
    assert var_mapping == {(2, 3): 6, (1, 2): 4, (1, 3): 5, (0, 1): 7}
    assert var_mapping.inverse == {6: (2, 3), 4: (1, 2), 5: (1, 3), 7: (0, 1)}

    var_mapping.inverse[7] = (0, 2)
    assert var_mapping == {(2, 3): 6, (1, 2): 4, (1, 3): 5, (0, 2): 7}
    assert var_mapping.inverse == {6: (2, 3), 4: (1, 2), 5: (1, 3), 7: (0, 2)}

    with pytest.raises(KeyAndValueDuplicationError):
        var_mapping[(2, 3)] = 4
    assert var_mapping == {(2, 3): 6, (1, 2): 4, (1, 3): 5, (0, 2): 7}

    with pytest.raises(ValueDuplicationError):
        var_mapping.inverse[9] = (2, 3)
    assert var_mapping == {(2, 3): 6, (1, 2): 4, (1, 3): 5, (0, 2): 7}

    with pytest.raises(ValueError, match="Invalid key type"):
        var_mapping[X] = 1

    with pytest.raises(ValueError, match="Invalid value type"):
        var_mapping[(1, 4)] = X

def test_del():
    """ test item deletion """
    var_mapping = VariableMapping(MAPPING_DICT)
    del var_mapping[(1, 2)]
    assert var_mapping == {(1, 3): 5}
    assert var_mapping.inverse == {5: (1, 3)}

    var_mapping = VariableMapping(MAPPING_DICT)
    del var_mapping.inverse[4]
    assert var_mapping == {(1, 3): 5}
    assert var_mapping.inverse == {5: (1, 3)}

def test_pop():
    """ test item pop """
    var_mapping = VariableMapping(MAPPING_DICT)
    assert var_mapping.pop((1, 2)) == 4
    assert var_mapping == {(1, 3): 5}
    assert var_mapping.inverse == {5: (1, 3)}

    var_mapping = VariableMapping(MAPPING_DICT)
    assert var_mapping.inverse.pop(4) == (1, 2)
    assert var_mapping == {(1, 3): 5}
    assert var_mapping.inverse == {5: (1, 3)}

    assert var_mapping.type_keys == tuple
    assert var_mapping.type_values == Integral
    assert var_mapping.inverse.pop(5) == (1, 3)
    assert not var_mapping
    assert not var_mapping.inverse
    assert not var_mapping.type_keys
    assert not var_mapping.type_values

    var_mapping = VariableMapping(MAPPING_DICT)
    item = var_mapping.popitem()
    assert item in [((1, 2), 4), ((1, 3), 5)]
    option1 = var_mapping == {(1, 3): 5} and var_mapping.inverse == {5: (1, 3)}
    option2 = var_mapping == {(1, 2): 4} and var_mapping.inverse == {4: (1, 2)}
    assert option1 or option2

    var_mapping = VariableMapping(MAPPING_DICT)
    item = var_mapping.inverse.popitem()
    assert item in [(4, (1, 2)), (5, (1, 3))]
    assert option1 or option2

def test_clear():
    """ test clearing of mapping """
    var_mapping = VariableMapping(MAPPING_DICT)
    assert var_mapping.type_keys == tuple
    assert var_mapping.type_values == Integral
    assert "type_values" in var_mapping.__dict__
    assert "type_keys" in var_mapping.__dict__

    var_mapping.clear()
    assert isinstance(var_mapping.inverse, VariableMapping)
    assert not var_mapping
    assert not var_mapping.inverse
    assert "type_values" not in var_mapping.__dict__
    assert "type_keys" not in var_mapping.__dict__
    assert "type_values" not in var_mapping.inverse.__dict__
    assert "type_keys" not in var_mapping.inverse.__dict__

def test_polynomial_use_case():
    """ test the variable mapping in relation to polynomials """
    tuple_poly = Polynomial({((X, 1),) : 1.0, ((X, 1), (X, 2)) : 3.0, () : 5.0})
    compact_poly = tuple_poly.compact()
    var_mapping = VariableMapping(tuple_poly.variables)

    decompact_poly = compact_poly.replace_variables(var_mapping)
    new_compact_poly = tuple_poly.replace_variables(var_mapping.inverse)
    assert decompact_poly == tuple_poly
    assert new_compact_poly == compact_poly

    var_num = compact_poly.get_variable_num()
    compact_solution = dict(enumerate(randint(0, 2, var_num)))
    decompact_solution = dict(zip(var_mapping.values(), compact_solution.values()))
    compact_value = compact_poly.evaluate(compact_solution)
    decompact_value = decompact_poly.evaluate(decompact_solution)
    assert compact_value == decompact_value

def test_repr():
    """ test string representation """
    assert repr(VariableMapping(MAPPING_DICT)) == "VariableMapping({(1, 2): 4, (1, 3): 5})"

def test_str():
    """ test string representation """
    poly = Polynomial({((Y, 1),) : 1, ((Y, 2),) : 1, ((Y, 3),) : 1})
    poly += Polynomial({((X, 1, 1),) : 1, ((X, 1, 2),) : 1, ((X, 2, 1),) : 1})
    var_mapping = VariableMapping(poly.variables)
    exp_str = "x0 <-> x_1_1\n" \
              "x1 <-> x_1_2\n" \
              "x2 <-> x_2_1\n" \
              "x3 <-> y_1\n" \
              "x4 <-> y_2\n" \
              "x5 <-> y_3"
    assert str(var_mapping) == exp_str

    exp_str = "x_1_1 <-> x0\n" \
              "x_1_2 <-> x1\n" \
              "x_2_1 <-> x2\n" \
              "y_1   <-> x3\n" \
              "y_2   <-> x4\n" \
              "y_3   <-> x5"
    assert str(var_mapping.inverse) == exp_str



# Copyright 2020 DLR-SC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" module for the IO of the PolyIsing """

import ast

from quark import PolyIsing, Polynomial
from . import add, hdf5, hdf5_attributes


INVERTED = "_inverted"


@add.self_method(PolyIsing)
def write_hdf5(polynomial, group):
    """
    write data in attributes and datasets in the opened group of an HDF5 file

    :param (PolyIsing) polynomial: the object whose data shall be stored
    :param (h5py.Group) group: the group to store the data in
    """
    hdf5.save_in(super(PolyIsing, polynomial), group)
    hdf5_attributes.write_attribute(group, INVERTED, polynomial)

@add.static_method(PolyIsing)
def read_hdf5(group):
    """
    read data from attributes and datasets of the opened group of an HDF5 file

    :param (h5py.Group) group: the group to read the data from
    :return: the data as keyword arguments
    """
    init_kwargs = hdf5.load_data_from(super(PolyIsing, PolyIsing), group)
    init_kwargs.update(inverted=hdf5_attributes.read_attribute(group, INVERTED))
    return init_kwargs


@add.self_method(PolyIsing)
def write_txt(polynomial, txt_file):
    """
    write data in the opened text file

    :param (PolyIsing) polynomial: the object whose data shall be stored
    :param (file) txt_file: the file to store the data in
    """
    txt_file.write(f"{INVERTED}={polynomial.is_inverted()}\n")
    txt_file.write(super(Polynomial, polynomial).__repr__().replace(", (", ",\n ("))

@add.static_method(PolyIsing)
def read_txt(lines):
    """
    read data from the opened text file

    :param (list) lines: the loaded lines of the text file to read the data from
    :return: the data as keyword arguments
    """
    inverted = ast.literal_eval(str(lines[0]).replace(f"{INVERTED}=", ""))
    poly_line = "".join(lines[1:]).replace("\n", "")
    poly_dict = ast.literal_eval(poly_line)
    return {"d": poly_dict, "inverted": inverted}
